from __future__ import annotations

from typing import TYPE_CHECKING, Generic, TypeVar

from narwhals._expression_parsing import ExprKind, ExprNode

if TYPE_CHECKING:
    from narwhals.expr import Expr
    from narwhals.typing import TimeUnit

ExprT = TypeVar("ExprT", bound="Expr")


class ExprDateTimeNamespace(Generic[ExprT]):
    def __init__(self, expr: ExprT) -> None:
        self._expr = expr

    def date(self) -> ExprT:
        """Extract the date from underlying DateTime representation.

        Raises:
            NotImplementedError: If pandas default backend is being used.

        Examples:
            >>> from datetime import datetime
            >>> import polars as pl
            >>> import narwhals as nw
            >>> df_native = pl.DataFrame(
            ...     {"a": [datetime(2012, 1, 7, 10), datetime(2027, 12, 13)]}
            ... )
            >>> df = nw.from_native(df_native)
            >>> df.select(nw.col("a").dt.date()).to_native()
            shape: (2, 1)
            ┌────────────┐
            │ a          │
            │ ---        │
            │ date       │
            ╞════════════╡
            │ 2012-01-07 │
            │ 2027-12-13 │
            └────────────┘
        """
        return self._expr._append_node(ExprNode(ExprKind.ELEMENTWISE, "dt.date"))

    def year(self) -> ExprT:
        """Extract year from underlying DateTime representation.

        Returns the year number in the calendar date.

        Examples:
            >>> from datetime import datetime
            >>> import pandas as pd
            >>> import narwhals as nw
            >>> df_native = pd.DataFrame(
            ...     {"a": [datetime(1978, 6, 1), datetime(2065, 1, 1)]}
            ... )
            >>> df = nw.from_native(df_native)
            >>> df.with_columns(nw.col("a").dt.year().alias("year"))
            ┌──────────────────┐
            |Narwhals DataFrame|
            |------------------|
            |           a  year|
            |0 1978-06-01  1978|
            |1 2065-01-01  2065|
            └──────────────────┘
        """
        return self._expr._append_node(ExprNode(ExprKind.ELEMENTWISE, "dt.year"))

    def month(self) -> ExprT:
        """Extract month from underlying DateTime representation.

        Returns the month number starting from 1. The return value ranges from 1 to 12.

        Examples:
            >>> from datetime import datetime
            >>> import pyarrow as pa
            >>> import narwhals as nw
            >>> df_native = pa.table({"a": [datetime(1978, 6, 1), datetime(2065, 1, 1)]})
            >>> df = nw.from_native(df_native)
            >>> df.with_columns(nw.col("a").dt.month().alias("month")).to_native()
            pyarrow.Table
            a: timestamp[us]
            month: int64
            ----
            a: [[1978-06-01 00:00:00.000000,2065-01-01 00:00:00.000000]]
            month: [[6,1]]
        """
        return self._expr._append_node(ExprNode(ExprKind.ELEMENTWISE, "dt.month"))

    def day(self) -> ExprT:
        """Extract day from underlying DateTime representation.

        Returns the day of month starting from 1. The return value ranges from 1 to 31. (The last day of month differs by months.)

        Examples:
            >>> from datetime import datetime
            >>> import pyarrow as pa
            >>> import narwhals as nw
            >>> df_native = pa.table({"a": [datetime(1978, 6, 1), datetime(2065, 1, 1)]})
            >>> df = nw.from_native(df_native)
            >>> df.with_columns(nw.col("a").dt.day().alias("day")).to_native()
            pyarrow.Table
            a: timestamp[us]
            day: int64
            ----
            a: [[1978-06-01 00:00:00.000000,2065-01-01 00:00:00.000000]]
            day: [[1,1]]
        """
        return self._expr._append_node(ExprNode(ExprKind.ELEMENTWISE, "dt.day"))

    def hour(self) -> ExprT:
        """Extract hour from underlying DateTime representation.

        Returns the hour number from 0 to 23.

        Examples:
            >>> from datetime import datetime
            >>> import polars as pl
            >>> import narwhals as nw
            >>> df_native = pl.DataFrame(
            ...     {"a": [datetime(1978, 1, 1, 1), datetime(2065, 1, 1, 10)]}
            ... )
            >>> df = nw.from_native(df_native)
            >>> df.with_columns(nw.col("a").dt.hour().alias("hour"))
            ┌──────────────────────────────┐
            |      Narwhals DataFrame      |
            |------------------------------|
            |shape: (2, 2)                 |
            |┌─────────────────────┬──────┐|
            |│ a                   ┆ hour │|
            |│ ---                 ┆ ---  │|
            |│ datetime[μs]        ┆ i8   │|
            |╞═════════════════════╪══════╡|
            |│ 1978-01-01 01:00:00 ┆ 1    │|
            |│ 2065-01-01 10:00:00 ┆ 10   │|
            |└─────────────────────┴──────┘|
            └──────────────────────────────┘
        """
        return self._expr._append_node(ExprNode(ExprKind.ELEMENTWISE, "dt.hour"))

    def minute(self) -> ExprT:
        """Extract minutes from underlying DateTime representation.

        Returns the minute number from 0 to 59.

        Examples:
            >>> from datetime import datetime
            >>> import pandas as pd
            >>> import narwhals as nw
            >>> df_native = pd.DataFrame(
            ...     {"a": [datetime(1978, 1, 1, 1, 1), datetime(2065, 1, 1, 10, 20)]}
            ... )
            >>> df = nw.from_native(df_native)
            >>> df.with_columns(nw.col("a").dt.minute().alias("minute")).to_native()
                                a  minute
            0 1978-01-01 01:01:00       1
            1 2065-01-01 10:20:00      20
        """
        return self._expr._append_node(ExprNode(ExprKind.ELEMENTWISE, "dt.minute"))

    def second(self) -> ExprT:
        """Extract seconds from underlying DateTime representation.

        Examples:
            >>> from datetime import datetime
            >>> import pyarrow as pa
            >>> import narwhals as nw
            >>> df_native = pa.table(
            ...     {
            ...         "a": [
            ...             datetime(1978, 1, 1, 1, 1, 1),
            ...             datetime(2065, 1, 1, 10, 20, 30),
            ...         ]
            ...     }
            ... )
            >>> df = nw.from_native(df_native)
            >>> df.with_columns(nw.col("a").dt.second().alias("second")).to_native()
            pyarrow.Table
            a: timestamp[us]
            second: int64
            ----
            a: [[1978-01-01 01:01:01.000000,2065-01-01 10:20:30.000000]]
            second: [[1,30]]
        """
        return self._expr._append_node(ExprNode(ExprKind.ELEMENTWISE, "dt.second"))

    def millisecond(self) -> ExprT:
        """Extract milliseconds from underlying DateTime representation.

        Examples:
            >>> from datetime import datetime
            >>> import pyarrow as pa
            >>> import narwhals as nw
            >>> df_native = pa.table(
            ...     {
            ...         "a": [
            ...             datetime(1978, 1, 1, 1, 1, 1, 0),
            ...             datetime(2065, 1, 1, 10, 20, 30, 67000),
            ...         ]
            ...     }
            ... )
            >>> df = nw.from_native(df_native)
            >>> df.with_columns(
            ...     nw.col("a").dt.millisecond().alias("millisecond")
            ... ).to_native()
            pyarrow.Table
            a: timestamp[us]
            millisecond: int64
            ----
            a: [[1978-01-01 01:01:01.000000,2065-01-01 10:20:30.067000]]
            millisecond: [[0,67]]
        """
        return self._expr._append_node(ExprNode(ExprKind.ELEMENTWISE, "dt.millisecond"))

    def microsecond(self) -> ExprT:
        """Extract microseconds from underlying DateTime representation.

        Examples:
            >>> from datetime import datetime
            >>> import pyarrow as pa
            >>> import narwhals as nw
            >>> df_native = pa.table(
            ...     {
            ...         "a": [
            ...             datetime(1978, 1, 1, 1, 1, 1, 0),
            ...             datetime(2065, 1, 1, 10, 20, 30, 67000),
            ...         ]
            ...     }
            ... )
            >>> df = nw.from_native(df_native)
            >>> df.with_columns(
            ...     nw.col("a").dt.microsecond().alias("microsecond")
            ... ).to_native()
            pyarrow.Table
            a: timestamp[us]
            microsecond: int64
            ----
            a: [[1978-01-01 01:01:01.000000,2065-01-01 10:20:30.067000]]
            microsecond: [[0,67000]]
        """
        return self._expr._append_node(ExprNode(ExprKind.ELEMENTWISE, "dt.microsecond"))

    def nanosecond(self) -> ExprT:
        """Extract Nanoseconds from underlying DateTime representation.

        Examples:
            >>> from datetime import datetime
            >>> import pyarrow as pa
            >>> import narwhals as nw
            >>> df_native = pa.table(
            ...     {
            ...         "a": [
            ...             datetime(1978, 1, 1, 1, 1, 1, 0),
            ...             datetime(2065, 1, 1, 10, 20, 30, 67000),
            ...         ]
            ...     }
            ... )
            >>> df = nw.from_native(df_native)
            >>> df.with_columns(
            ...     nw.col("a").dt.nanosecond().alias("nanosecond")
            ... ).to_native()
            pyarrow.Table
            a: timestamp[us]
            nanosecond: int64
            ----
            a: [[1978-01-01 01:01:01.000000,2065-01-01 10:20:30.067000]]
            nanosecond: [[0,67000000]]
        """
        return self._expr._append_node(ExprNode(ExprKind.ELEMENTWISE, "dt.nanosecond"))

    def ordinal_day(self) -> ExprT:
        """Get ordinal day.

        Examples:
            >>> from datetime import datetime
            >>> import pandas as pd
            >>> import narwhals as nw
            >>> df_native = pd.DataFrame(
            ...     {"a": [datetime(2020, 1, 1), datetime(2020, 8, 3)]}
            ... )
            >>> df = nw.from_native(df_native)
            >>> df.with_columns(a_ordinal_day=nw.col("a").dt.ordinal_day())
            ┌───────────────────────────┐
            |    Narwhals DataFrame     |
            |---------------------------|
            |           a  a_ordinal_day|
            |0 2020-01-01              1|
            |1 2020-08-03            216|
            └───────────────────────────┘
        """
        return self._expr._append_node(ExprNode(ExprKind.ELEMENTWISE, "dt.ordinal_day"))

    def weekday(self) -> ExprT:
        """Extract the week day from the underlying Date representation.

        Note that Monday = 1 and Sunday = 7.

        Examples:
            >>> from datetime import datetime
            >>> import pandas as pd
            >>> import narwhals as nw
            >>> df_native = pd.DataFrame(
            ...     {"a": [datetime(2020, 1, 1), datetime(2020, 8, 3)]}
            ... )
            >>> df = nw.from_native(df_native)
            >>> df.with_columns(a_week_day=nw.col("a").dt.weekday())
            ┌────────────────────────┐
            |   Narwhals DataFrame   |
            |------------------------|
            |           a  a_week_day|
            |0 2020-01-01           3|
            |1 2020-08-03           1|
            └────────────────────────┘
        """
        return self._expr._append_node(ExprNode(ExprKind.ELEMENTWISE, "dt.weekday"))

    def total_minutes(self) -> ExprT:
        """Get total minutes.

        Notes:
            The function outputs the total minutes in the int dtype by default,
            however, pandas may change the dtype to float when there are missing values,
            consider using `fill_null()` and `cast` in this case.

        Examples:
            >>> from datetime import timedelta
            >>> import polars as pl
            >>> import narwhals as nw
            >>> df_native = pl.DataFrame(
            ...     {"a": [timedelta(minutes=10), timedelta(minutes=20, seconds=40)]}
            ... )
            >>> df = nw.from_native(df_native)
            >>> df.with_columns(
            ...     a_total_minutes=nw.col("a").dt.total_minutes()
            ... ).to_native()
            shape: (2, 2)
            ┌──────────────┬─────────────────┐
            │ a            ┆ a_total_minutes │
            │ ---          ┆ ---             │
            │ duration[μs] ┆ i64             │
            ╞══════════════╪═════════════════╡
            │ 10m          ┆ 10              │
            │ 20m 40s      ┆ 20              │
            └──────────────┴─────────────────┘
        """
        return self._expr._append_node(ExprNode(ExprKind.ELEMENTWISE, "dt.total_minutes"))

    def total_seconds(self) -> ExprT:
        """Get total seconds.

        Notes:
            The function outputs the total seconds in the int dtype by default,
            however, pandas may change the dtype to float when there are missing values,
            consider using `fill_null()` and `cast` in this case.

        Examples:
            >>> from datetime import timedelta
            >>> import polars as pl
            >>> import narwhals as nw
            >>> df_native = pl.DataFrame(
            ...     {"a": [timedelta(seconds=10), timedelta(seconds=20, milliseconds=40)]}
            ... )
            >>> df = nw.from_native(df_native)
            >>> df.with_columns(
            ...     a_total_seconds=nw.col("a").dt.total_seconds()
            ... ).to_native()
            shape: (2, 2)
            ┌──────────────┬─────────────────┐
            │ a            ┆ a_total_seconds │
            │ ---          ┆ ---             │
            │ duration[μs] ┆ i64             │
            ╞══════════════╪═════════════════╡
            │ 10s          ┆ 10              │
            │ 20s 40ms     ┆ 20              │
            └──────────────┴─────────────────┘
        """
        return self._expr._append_node(ExprNode(ExprKind.ELEMENTWISE, "dt.total_seconds"))

    def total_milliseconds(self) -> ExprT:
        """Get total milliseconds.

        Notes:
            The function outputs the total milliseconds in the int dtype by default,
            however, pandas may change the dtype to float when there are missing values,
            consider using `fill_null()` and `cast` in this case.

        Examples:
            >>> from datetime import timedelta
            >>> import polars as pl
            >>> import narwhals as nw
            >>> df_native = pl.DataFrame(
            ...     {
            ...         "a": [
            ...             timedelta(milliseconds=10),
            ...             timedelta(milliseconds=20, microseconds=40),
            ...         ]
            ...     }
            ... )
            >>> df = nw.from_native(df_native)
            >>> df.with_columns(
            ...     a_total_milliseconds=nw.col("a").dt.total_milliseconds()
            ... ).to_native()
            shape: (2, 2)
            ┌──────────────┬──────────────────────┐
            │ a            ┆ a_total_milliseconds │
            │ ---          ┆ ---                  │
            │ duration[μs] ┆ i64                  │
            ╞══════════════╪══════════════════════╡
            │ 10ms         ┆ 10                   │
            │ 20040µs      ┆ 20                   │
            └──────────────┴──────────────────────┘
        """
        return self._expr._append_node(
            ExprNode(ExprKind.ELEMENTWISE, "dt.total_milliseconds")
        )

    def total_microseconds(self) -> ExprT:
        """Get total microseconds.

        Notes:
            The function outputs the total microseconds in the int dtype by default,
            however, pandas may change the dtype to float when there are missing values,
            consider using `fill_null()` and `cast` in this case.

        Examples:
            >>> from datetime import timedelta
            >>> import pyarrow as pa
            >>> import narwhals as nw
            >>> df_native = pa.table(
            ...     {
            ...         "a": [
            ...             timedelta(microseconds=10),
            ...             timedelta(milliseconds=1, microseconds=200),
            ...         ]
            ...     }
            ... )
            >>> df = nw.from_native(df_native)
            >>> df.with_columns(
            ...     a_total_microseconds=nw.col("a").dt.total_microseconds()
            ... ).to_native()
            pyarrow.Table
            a: duration[us]
            a_total_microseconds: int64
            ----
            a: [[10,1200]]
            a_total_microseconds: [[10,1200]]
        """
        return self._expr._append_node(
            ExprNode(ExprKind.ELEMENTWISE, "dt.total_microseconds")
        )

    def total_nanoseconds(self) -> ExprT:
        """Get total nanoseconds.

        Notes:
            The function outputs the total nanoseconds in the int dtype by default,
            however, pandas may change the dtype to float when there are missing values,
            consider using `fill_null()` and `cast` in this case.

        Examples:
            >>> from datetime import timedelta
            >>> import pandas as pd
            >>> import narwhals as nw
            >>> df_native = pd.DataFrame(
            ...     {
            ...         "a": pd.to_datetime(
            ...             [
            ...                 "2024-01-01 00:00:00.000000001",
            ...                 "2024-01-01 00:00:00.000000002",
            ...             ]
            ...         )
            ...     }
            ... )
            >>> df = nw.from_native(df_native)
            >>> df.with_columns(
            ...     a_diff_total_nanoseconds=nw.col("a").diff().dt.total_nanoseconds()
            ... ).to_native()
                                          a  a_diff_total_nanoseconds
            0 2024-01-01 00:00:00.000000001                       NaN
            1 2024-01-01 00:00:00.000000002                       1.0
        """
        return self._expr._append_node(
            ExprNode(ExprKind.ELEMENTWISE, "dt.total_nanoseconds")
        )

    def to_string(self, format: str) -> ExprT:
        """Convert a Date/Time/Datetime column into a String column with the given format.

        Arguments:
            format: Format to format temporal column with.

        Notes:
            Unfortunately, different libraries interpret format directives a bit
            differently.

            - Chrono, the library used by Polars, uses `"%.f"` for fractional seconds,
              whereas pandas and Python stdlib use `".%f"`.
            - PyArrow interprets `"%S"` as "seconds, including fractional seconds"
              whereas most other tools interpret it as "just seconds, as 2 digits".
            ---
            Therefore, we make the following adjustments.

            - for pandas-like libraries, we replace `"%S.%f"` with `"%S%.f"`.
            - for PyArrow, we replace `"%S.%f"` with `"%S"`.
            ---
            Workarounds like these don't make us happy, and we try to avoid them as
            much as possible, but here we feel like it's the best compromise.

            If you just want to format a date/datetime Series as a local datetime
            string, and have it work as consistently as possible across libraries,
            we suggest using:

            - `"%Y-%m-%dT%H:%M:%S%.f"` for datetimes
            - `"%Y-%m-%d"` for dates
            ---
            Though note that, even then, different tools may return a different number
            of trailing zeros. Nonetheless, this is probably consistent enough for
            most applications.

            If you have an application where this is not enough, please open an issue
            and let us know.

        Examples:
            >>> from datetime import datetime
            >>> import polars as pl
            >>> import narwhals as nw
            >>> df_native = pl.DataFrame(
            ...     {"a": [datetime(2020, 3, 1), datetime(2020, 5, 1)]}
            ... )
            >>> df = nw.from_native(df_native)
            >>> df.select(nw.col("a").dt.to_string("%Y/%m/%d %H:%M:%S"))
            ┌───────────────────────┐
            |  Narwhals DataFrame   |
            |-----------------------|
            |shape: (2, 1)          |
            |┌─────────────────────┐|
            |│ a                   │|
            |│ ---                 │|
            |│ str                 │|
            |╞═════════════════════╡|
            |│ 2020/03/01 00:00:00 │|
            |│ 2020/05/01 00:00:00 │|
            |└─────────────────────┘|
            └───────────────────────┘
        """
        return self._expr._append_node(
            ExprNode(ExprKind.ELEMENTWISE, "dt.to_string", format=format)
        )

    def replace_time_zone(self, time_zone: str | None) -> ExprT:
        """Replace time zone.

        Arguments:
            time_zone: Target time zone.

        Examples:
            >>> from datetime import datetime, timezone
            >>> import pandas as pd
            >>> import narwhals as nw
            >>> df_native = pd.DataFrame(
            ...     {
            ...         "a": [
            ...             datetime(2024, 1, 1, tzinfo=timezone.utc),
            ...             datetime(2024, 1, 2, tzinfo=timezone.utc),
            ...         ]
            ...     }
            ... )
            >>> df = nw.from_native(df_native)
            >>> df.select(nw.col("a").dt.replace_time_zone("Asia/Kathmandu")).to_native()
                                      a
            0 2024-01-01 00:00:00+05:45
            1 2024-01-02 00:00:00+05:45
        """
        return self._expr._append_node(
            ExprNode(ExprKind.ELEMENTWISE, "dt.replace_time_zone", time_zone=time_zone)
        )

    def convert_time_zone(self, time_zone: str) -> ExprT:
        """Convert to a new time zone.

        If converting from a time-zone-naive column, then conversion happens
        as if converting from UTC.

        Arguments:
            time_zone: Target time zone.

        Examples:
            >>> from datetime import datetime, timezone
            >>> import pandas as pd
            >>> import narwhals as nw
            >>> df_native = pd.DataFrame(
            ...     {
            ...         "a": [
            ...             datetime(2024, 1, 1, tzinfo=timezone.utc),
            ...             datetime(2024, 1, 2, tzinfo=timezone.utc),
            ...         ]
            ...     }
            ... )
            >>> df = nw.from_native(df_native)
            >>> df.select(nw.col("a").dt.convert_time_zone("Asia/Kathmandu")).to_native()
                                      a
            0 2024-01-01 05:45:00+05:45
            1 2024-01-02 05:45:00+05:45
        """
        if time_zone is None:
            msg = "Target `time_zone` cannot be `None` in `convert_time_zone`. Please use `replace_time_zone(None)` if you want to remove the time zone."
            raise TypeError(msg)
        return self._expr._append_node(
            ExprNode(ExprKind.ELEMENTWISE, "dt.convert_time_zone", time_zone=time_zone)
        )

    def timestamp(self, time_unit: TimeUnit = "us") -> ExprT:
        """Return a timestamp in the given time unit.

        Arguments:
            time_unit: One of
                - 'ns': nanosecond.
                - 'us': microsecond.
                - 'ms': millisecond.

        Examples:
            >>> from datetime import date
            >>> import polars as pl
            >>> import narwhals as nw
            >>> df_native = pl.DataFrame({"date": [date(2001, 1, 1), None]})
            >>> df = nw.from_native(df_native)
            >>> df.with_columns(nw.col("date").dt.timestamp("ms").alias("timestamp_ms"))
            ┌─────────────────────────────┐
            |     Narwhals DataFrame      |
            |-----------------------------|
            |shape: (2, 2)                |
            |┌────────────┬──────────────┐|
            |│ date       ┆ timestamp_ms │|
            |│ ---        ┆ ---          │|
            |│ date       ┆ i64          │|
            |╞════════════╪══════════════╡|
            |│ 2001-01-01 ┆ 978307200000 │|
            |│ null       ┆ null         │|
            |└────────────┴──────────────┘|
            └─────────────────────────────┘
        """
        if time_unit not in {"ns", "us", "ms"}:
            msg = (
                "invalid `time_unit`"
                f"\n\nExpected one of {{'ns', 'us', 'ms'}}, got {time_unit!r}."
            )
            raise ValueError(msg)
        return self._expr._append_node(
            ExprNode(ExprKind.ELEMENTWISE, "dt.timestamp", time_unit=time_unit)
        )

    def truncate(self, every: str) -> ExprT:
        """Divide the date/datetime range into buckets.

        Arguments:
            every: Length of bucket. Must be of form `<multiple><unit>`,
                where `multiple` is a positive integer and `unit` is one of

                - 'ns': nanosecond.
                - 'us': microsecond.
                - 'ms': millisecond.
                - 's': second.
                - 'm': minute.
                - 'h': hour.
                - 'd': day.
                - 'mo': month.
                - 'q': quarter.
                - 'y': year.

        Examples:
            >>> from datetime import datetime
            >>> import polars as pl
            >>> import narwhals as nw
            >>> df_native = pl.DataFrame({"datetime": [datetime(2021, 3, 1, 12, 34)]})
            >>> df = nw.from_native(df_native)
            >>> df.with_columns(
            ...     nw.col("datetime").dt.truncate("1h").alias("datetime_trunc")
            ... )
            ┌─────────────────────────────────────────────┐
            |             Narwhals DataFrame              |
            |---------------------------------------------|
            |shape: (1, 2)                                |
            |┌─────────────────────┬─────────────────────┐|
            |│ datetime            ┆ datetime_trunc      │|
            |│ ---                 ┆ ---                 │|
            |│ datetime[μs]        ┆ datetime[μs]        │|
            |╞═════════════════════╪═════════════════════╡|
            |│ 2021-03-01 12:34:00 ┆ 2021-03-01 12:00:00 │|
            |└─────────────────────┴─────────────────────┘|
            └─────────────────────────────────────────────┘
        """
        return self._expr._append_node(
            ExprNode(ExprKind.ELEMENTWISE, "dt.truncate", every=every)
        )

    def offset_by(self, by: str) -> ExprT:
        """Offset this date by a relative time offset.

        Arguments:
            by: The offset. Must be of form `<multiple><unit>`,
                where `multiple` is a positive integer and `unit` is one of

                - 'ns': nanosecond.
                - 'us': microsecond.
                - 'ms': millisecond.
                - 's': second.
                - 'm': minute.
                - 'h': hour.
                - 'd': day.
                - 'mo': month.
                - 'q': quarter.
                - 'y': year.

        Examples:
            >>> from datetime import datetime
            >>> import polars as pl
            >>> import narwhals as nw
            >>> df_native = pl.DataFrame({"datetime": [datetime(2021, 3, 1, 12, 34)]})
            >>> df = nw.from_native(df_native)
            >>> df.with_columns(
            ...     nw.col("datetime").dt.offset_by("1h").alias("datetime_offset_by_1h")
            ... )
            ┌───────────────────────────────────────────────┐
            |              Narwhals DataFrame               |
            |-----------------------------------------------|
            |shape: (1, 2)                                  |
            |┌─────────────────────┬───────────────────────┐|
            |│ datetime            ┆ datetime_offset_by_1h │|
            |│ ---                 ┆ ---                   │|
            |│ datetime[μs]        ┆ datetime[μs]          │|
            |╞═════════════════════╪═══════════════════════╡|
            |│ 2021-03-01 12:34:00 ┆ 2021-03-01 13:34:00   │|
            |└─────────────────────┴───────────────────────┘|
            └───────────────────────────────────────────────┘
        """
        return self._expr._append_node(
            ExprNode(ExprKind.ELEMENTWISE, "dt.offset_by", by=by)
        )



"""Check that `output.txt` doesn't contain the string "exited with errors".

If it does, exit with status 1.

This is just used in CI.
"""

from __future__ import annotations

import sys
from pathlib import Path

if "exited with errors" in Path("output.txt").read_text("utf-8"):
    sys.exit(1)
sys.exit(0)



from __future__ import annotations

from contextlib import nullcontext as does_not_raise

import pytest

import narwhals as nw
from tests.utils import DUCKDB_VERSION, Constructor, ConstructorEager, assert_equal_data


def test_unary(constructor: Constructor, request: pytest.FixtureRequest) -> None:
    if "ibis" in str(constructor):
        request.applymarker(pytest.mark.xfail)
    if "duckdb" in str(constructor) and DUCKDB_VERSION < (1, 3):
        pytest.skip()

    data = {"a": [1, 3, 2], "b": [4, 4, 6], "c": [7.0, 8.0, None], "z": [7.0, 8.0, 9.0]}
    result = nw.from_native(constructor(data)).select(
        a_mean=nw.col("a").mean(),
        a_median=nw.col("a").median(),
        a_sum=nw.col("a").sum(),
        a_skew=nw.col("a").skew(),
        b_nunique=nw.col("b").n_unique(),
        b_skew=nw.col("b").skew(),
        c_nunique=nw.col("c").n_unique(),
        z_min=nw.col("z").min(),
        z_max=nw.col("z").max(),
    )
    expected = {
        "a_mean": [2],
        "a_median": [2],
        "a_sum": [6],
        "a_skew": [0.0],
        "b_nunique": [2],
        "b_skew": [0.7071067811865465],
        "c_nunique": [3],
        "z_min": [7],
        "z_max": [9],
    }
    assert_equal_data(result, expected)


def test_unary_series(constructor_eager: ConstructorEager) -> None:
    data = {"a": [1, 3, 2], "b": [4, 4, 6], "c": [7.0, 8.0, None], "z": [7.0, 8.0, 9.0]}
    df = nw.from_native(constructor_eager(data), eager_only=True)
    result = {
        "a_mean": [df["a"].mean()],
        "a_median": [df["a"].median()],
        "a_sum": [df["a"].sum()],
        "a_skew": [df["a"].skew()],
        "b_nunique": [df["b"].n_unique()],
        "b_skew": [df["b"].skew()],
        "c_nunique": [df["c"].n_unique()],
        "c_skew": [df["c"].skew()],
        "z_min": [df["z"].min()],
        "z_max": [df["z"].max()],
    }
    expected = {
        "a_mean": [2.0],
        "a_median": [2],
        "a_sum": [6],
        "a_skew": [0.0],
        "b_nunique": [2],
        "b_skew": [0.7071067811865465],
        "c_nunique": [3],
        "c_skew": [0.0],
        "z_min": [7.0],
        "z_max": [9.0],
    }
    assert_equal_data(result, expected)


def test_unary_two_elements(
    constructor: Constructor, request: pytest.FixtureRequest
) -> None:
    if "ibis" in str(constructor):
        request.applymarker(pytest.mark.xfail)
    if "duckdb" in str(constructor) and DUCKDB_VERSION < (1, 3):
        pytest.skip()
    data = {"a": [1, 2], "b": [2, 10], "c": [2.0, None]}
    result = nw.from_native(constructor(data)).select(
        a_nunique=nw.col("a").n_unique(),
        a_skew=nw.col("a").skew(),
        b_nunique=nw.col("b").n_unique(),
        b_skew=nw.col("b").skew(),
        c_nunique=nw.col("c").n_unique(),
        c_skew=nw.col("c").skew(),
    )
    expected = {
        "a_nunique": [2],
        "a_skew": [0.0],
        "b_nunique": [2],
        "b_skew": [0.0],
        "c_nunique": [2],
        "c_skew": [None],
    }
    assert_equal_data(result, expected)


def test_unary_two_elements_series(constructor_eager: ConstructorEager) -> None:
    data = {"a": [1, 2], "b": [2, 10], "c": [2.0, None]}
    df = nw.from_native(constructor_eager(data), eager_only=True)
    result = {
        "a_nunique": [df["a"].n_unique()],
        "a_skew": [df["a"].skew()],
        "b_nunique": [df["b"].n_unique()],
        "b_skew": [df["b"].skew()],
        "c_nunique": [df["c"].n_unique()],
        "c_skew": [df["c"].skew()],
    }
    expected = {
        "a_nunique": [2],
        "a_skew": [0.0],
        "b_nunique": [2],
        "b_skew": [0.0],
        "c_nunique": [2],
        "c_skew": [None],
    }
    assert_equal_data(result, expected)


def test_unary_one_element(
    constructor: Constructor, request: pytest.FixtureRequest
) -> None:
    if "pyspark" in str(constructor) and "sqlframe" not in str(constructor):
        request.applymarker(pytest.mark.xfail)
    if "ibis" in str(constructor):
        request.applymarker(pytest.mark.xfail)
    if "duckdb" in str(constructor) and DUCKDB_VERSION < (1, 3):
        pytest.skip()
    data = {"a": [1], "b": [2], "c": [None]}
    # Dask runs into a divide by zero RuntimeWarning for 1 element skew.
    context = (
        pytest.warns(RuntimeWarning, match="invalid value encountered in scalar divide")
        if "dask" in str(constructor)
        else does_not_raise()
    )

    result = (
        nw.from_native(constructor(data))
        .with_columns(nw.col("c").cast(nw.Float64))
        .select(
            a_nunique=nw.col("a").n_unique(),
            a_skew=nw.col("a").skew(),
            b_nunique=nw.col("b").n_unique(),
            b_skew=nw.col("b").skew(),
            c_nunique=nw.col("c").n_unique(),
            c_skew=nw.col("c").skew(),
        )
    )
    expected = {
        "a_nunique": [1],
        "a_skew": [None],
        "b_nunique": [1],
        "b_skew": [None],
        "c_nunique": [1],
        "c_skew": [None],
    }
    with context:
        assert_equal_data(result, expected)


def test_unary_one_element_series(constructor_eager: ConstructorEager) -> None:
    data = {"a": [1], "b": [2], "c": [None]}
    df = nw.from_native(constructor_eager(data), eager_only=True)
    result = {
        "a_nunique": [df["a"].n_unique()],
        "a_skew": [df["a"].skew()],
        "b_nunique": [df["b"].n_unique()],
        "b_skew": [df["b"].skew()],
        "c_nunique": [df["c"].n_unique()],
        "c_skew": [df["c"].skew()],
    }
    expected = {
        "a_nunique": [1],
        "a_skew": [None],
        "b_nunique": [1],
        "b_skew": [None],
        "c_nunique": [1],
        "c_skew": [None],
    }
    assert_equal_data(result, expected)



from __future__ import annotations

from typing import TYPE_CHECKING

import narwhals as nw

if TYPE_CHECKING:
    from narwhals.typing import FrameT


def query(lineitem: FrameT, nation: FrameT, orders: FrameT, supplier: FrameT) -> FrameT:
    var1 = "SAUDI ARABIA"

    q1 = (
        lineitem.group_by("l_orderkey")
        .agg(nw.len().alias("n_supp_by_order"))
        .filter(nw.col("n_supp_by_order") > 1)
        .join(
            lineitem.filter(nw.col("l_receiptdate") > nw.col("l_commitdate")),  # pyright: ignore[reportArgumentType]
            left_on="l_orderkey",
            right_on="l_orderkey",
        )
    )

    return (
        q1.group_by("l_orderkey")
        .agg(nw.len().alias("n_supp_by_order"))
        .join(
            q1,  # pyright: ignore[reportArgumentType]
            left_on="l_orderkey",
            right_on="l_orderkey",
        )
        .join(supplier, left_on="l_suppkey", right_on="s_suppkey")
        .join(nation, left_on="s_nationkey", right_on="n_nationkey")
        .join(orders, left_on="l_orderkey", right_on="o_orderkey")
        .filter(nw.col("n_supp_by_order") == 1)
        .filter(nw.col("n_name") == var1)
        .filter(nw.col("o_orderstatus") == "F")
        .group_by("s_name")
        .agg(nw.len().alias("numwait"))
        .sort(by=["numwait", "s_name"], descending=[True, False])
        .head(100)
    )



from __future__ import annotations

from collections.abc import Iterator, Mapping, Sequence, Sized
from typing import TYPE_CHECKING, Any, Generic, Literal, TypeVar, cast, overload

import polars as pl

from narwhals._polars.namespace import PolarsNamespace
from narwhals._polars.series import PolarsSeries
from narwhals._polars.utils import (
    FROM_DICTS_ACCEPTS_MAPPINGS,
    catch_polars_exception,
    extract_args_kwargs,
    narwhals_to_native_dtype,
    native_to_narwhals_dtype,
)
from narwhals._utils import (
    Implementation,
    _into_arrow_table,
    convert_str_slice_to_int_slice,
    generate_temporary_column_name,
    is_boolean_selector,
    is_compliant_series,
    is_index_selector,
    is_range,
    is_sequence_like,
    is_slice_index,
    is_slice_none,
    parse_columns_to_drop,
    requires,
)
from narwhals.dependencies import is_numpy_array_1d
from narwhals.exceptions import ColumnNotFoundError

if TYPE_CHECKING:
    from collections.abc import Iterable
    from types import ModuleType
    from typing import Callable

    import pandas as pd
    import pyarrow as pa
    from typing_extensions import Self, TypeAlias, TypeIs

    from narwhals._compliant.typing import CompliantDataFrameAny, CompliantLazyFrameAny
    from narwhals._polars.expr import PolarsExpr
    from narwhals._polars.group_by import PolarsGroupBy, PolarsLazyGroupBy
    from narwhals._spark_like.utils import SparkSession
    from narwhals._translate import IntoArrowTable
    from narwhals._typing import _EagerAllowedImpl, _LazyAllowedImpl
    from narwhals._utils import Version, _LimitedContext
    from narwhals.dataframe import DataFrame, LazyFrame
    from narwhals.dtypes import DType
    from narwhals.typing import (
        IntoSchema,
        JoinStrategy,
        MultiColSelector,
        MultiIndexSelector,
        PivotAgg,
        SingleIndexSelector,
        UniqueKeepStrategy,
        _2DArray,
    )

    T = TypeVar("T")
    R = TypeVar("R")

Method: TypeAlias = "Callable[..., R]"
"""Generic alias representing all methods implemented via `__getattr__`.

Where `R` is the return type.
"""

# DataFrame methods where PolarsDataFrame just defers to Polars.DataFrame directly.
INHERITED_METHODS = frozenset(
    [
        "clone",
        "drop_nulls",
        "estimated_size",
        "explode",
        "filter",
        "gather_every",
        "head",
        "is_unique",
        "item",
        "iter_rows",
        "join_asof",
        "rename",
        "row",
        "rows",
        "sample",
        "select",
        "sink_parquet",
        "sort",
        "tail",
        "to_arrow",
        "to_pandas",
        "with_columns",
        "write_csv",
        "write_parquet",
    ]
)

NativePolarsFrame = TypeVar("NativePolarsFrame", pl.DataFrame, pl.LazyFrame)


class PolarsBaseFrame(Generic[NativePolarsFrame]):
    drop_nulls: Method[Self]
    explode: Method[Self]
    filter: Method[Self]
    gather_every: Method[Self]
    head: Method[Self]
    join_asof: Method[Self]
    rename: Method[Self]
    select: Method[Self]
    sort: Method[Self]
    tail: Method[Self]
    with_columns: Method[Self]

    _native_frame: NativePolarsFrame
    _implementation = Implementation.POLARS
    _version: Version

    def __init__(
        self,
        df: NativePolarsFrame,
        *,
        version: Version,
        validate_backend_version: bool = False,
    ) -> None:
        self._native_frame = df
        self._version = version
        if validate_backend_version:
            self._validate_backend_version()

    def _validate_backend_version(self) -> None:
        """Raise if installed version below `nw._utils.MIN_VERSIONS`.

        **Only use this when moving between backends.**
        Otherwise, the validation will have taken place already.
        """
        _ = self._implementation._backend_version()

    @property
    def _backend_version(self) -> tuple[int, ...]:
        return self._implementation._backend_version()

    @property
    def native(self) -> NativePolarsFrame:
        return self._native_frame

    @property
    def columns(self) -> list[str]:
        return self.native.columns

    def __narwhals_namespace__(self) -> PolarsNamespace:
        return PolarsNamespace(version=self._version)

    def __native_namespace__(self) -> ModuleType:
        if self._implementation is Implementation.POLARS:
            return self._implementation.to_native_namespace()

        msg = f"Expected polars, got: {type(self._implementation)}"  # pragma: no cover
        raise AssertionError(msg)

    def _with_native(self, df: NativePolarsFrame) -> Self:
        return self.__class__(df, version=self._version)

    def _with_version(self, version: Version) -> Self:
        return self.__class__(self.native, version=version)

    @classmethod
    def from_native(cls, data: NativePolarsFrame, /, *, context: _LimitedContext) -> Self:
        return cls(data, version=context._version)

    def simple_select(self, *column_names: str) -> Self:
        return self._with_native(self.native.select(*column_names))

    def aggregate(self, *exprs: Any) -> Self:
        return self.select(*exprs)

    def unique(
        self,
        subset: Sequence[str] | None,
        *,
        keep: UniqueKeepStrategy,
        maintain_order: bool | None = None,
        order_by: Sequence[str] | None = None,
    ) -> Self:
        if order_by and maintain_order:
            token = generate_temporary_column_name(8, self.columns, prefix="row_index_")
            res = (
                self.native.with_row_index(token)
                .sort(order_by, nulls_last=False)
                .unique(subset or self.columns, keep=keep)
                .sort(token)
                .drop(token)
            )
        elif order_by:
            res = self.native.sort(order_by).unique(subset, keep=keep)
        else:
            res = self.native.unique(
                subset, keep=keep, maintain_order=maintain_order or False
            )
        return self._with_native(res)

    @property
    def schema(self) -> dict[str, DType]:
        return self.collect_schema()

    def join(
        self,
        other: PolarsBaseFrame[NativePolarsFrame],
        *,
        how: JoinStrategy,
        left_on: Sequence[str] | None,
        right_on: Sequence[str] | None,
        suffix: str,
    ) -> Self:
        how_native = (
            "outer" if (self._backend_version < (0, 20, 29) and how == "full") else how
        )
        return self._with_native(
            self.native.join(
                other=other.native,
                how=how_native,  # type: ignore[arg-type]
                left_on=left_on,
                right_on=right_on,
                suffix=suffix,
            )
        )

    def top_k(
        self, k: int, *, by: str | Iterable[str], reverse: bool | Sequence[bool]
    ) -> Self:
        if self._backend_version < (1, 0, 0):
            return self._with_native(
                self.native.top_k(
                    k=k,
                    by=by,
                    descending=reverse,  # type: ignore[call-arg]
                )
            )
        return self._with_native(self.native.top_k(k=k, by=by, reverse=reverse))

    def unpivot(
        self,
        on: Sequence[str] | None,
        index: Sequence[str] | None,
        variable_name: str,
        value_name: str,
    ) -> Self:
        if self._backend_version < (1, 0, 0):
            return self._with_native(
                self.native.melt(
                    id_vars=index,
                    value_vars=on,
                    variable_name=variable_name,
                    value_name=value_name,
                )
            )
        return self._with_native(
            self.native.unpivot(
                on=on, index=index, variable_name=variable_name, value_name=value_name
            )
        )

    def collect_schema(self) -> dict[str, DType]:
        df = self.native
        schema = df.schema if self._backend_version < (1,) else df.collect_schema()
        return {
            name: native_to_narwhals_dtype(dtype, self._version)
            for name, dtype in schema.items()
        }

    def with_row_index(self, name: str, order_by: Sequence[str] | None) -> Self:
        frame = self.native
        if order_by is None:
            result = frame.with_row_index(name)
        else:
            result = frame.select(
                pl.int_range(pl.len()).over(order_by=order_by).alias(name), pl.all()
            )

        return self._with_native(result)


class PolarsDataFrame(PolarsBaseFrame[pl.DataFrame]):
    clone: Method[Self]
    collect: Method[CompliantDataFrameAny]
    estimated_size: Method[int | float]
    gather_every: Method[Self]
    iter_rows: Method[Iterator[tuple[Any, ...]] | Iterator[Mapping[str, Any]]]
    is_unique: Method[PolarsSeries]
    row: Method[tuple[Any, ...]]
    rows: Method[Sequence[tuple[Any, ...]] | Sequence[Mapping[str, Any]]]
    sample: Method[Self]
    to_arrow: Method[pa.Table]
    to_pandas: Method[pd.DataFrame]
    # NOTE: `write_csv` requires an `@overload` for `str | None`
    # Can't do that here 😟
    write_csv: Method[Any]
    write_parquet: Method[None]

    @classmethod
    def from_arrow(cls, data: IntoArrowTable, /, *, context: _LimitedContext) -> Self:
        if context._implementation._backend_version() >= (1, 3):
            native = pl.DataFrame(data)
        else:  # pragma: no cover
            native = cast("pl.DataFrame", pl.from_arrow(_into_arrow_table(data, context)))
        return cls.from_native(native, context=context)

    @classmethod
    def from_dict(
        cls,
        data: Mapping[str, Any],
        /,
        *,
        context: _LimitedContext,
        schema: IntoSchema | Mapping[str, DType | None] | None,
    ) -> Self:
        pl_schema = (
            {
                key: narwhals_to_native_dtype(dtype, context._version)
                if dtype is not None
                else None
                for (key, dtype) in schema.items()
            }
            if schema
            else None
        )
        return cls.from_native(pl.from_dict(data, pl_schema), context=context)

    @classmethod
    def from_dicts(
        cls,
        data: Sequence[Mapping[str, Any]],
        /,
        *,
        context: _LimitedContext,
        schema: IntoSchema | Mapping[str, DType | None] | None,
    ) -> Self:
        pl_schema = (
            {
                key: narwhals_to_native_dtype(dtype, context._version)
                if dtype is not None
                else None
                for (key, dtype) in schema.items()
            }
            if schema
            else None
        )
        if not data:
            native = pl.DataFrame(schema=pl_schema)
        elif FROM_DICTS_ACCEPTS_MAPPINGS or isinstance(data[0], dict):
            native = pl.from_dicts(data, pl_schema)
        else:  # pragma: no cover
            columns = pl_schema or tuple(data[0])
            native = pl.DataFrame(
                (tuple(row.values()) for row in data), schema=columns, orient="row"
            )

        return cls.from_native(native, context=context)

    @staticmethod
    def _is_native(obj: pl.DataFrame | Any) -> TypeIs[pl.DataFrame]:
        return isinstance(obj, pl.DataFrame)

    @classmethod
    def from_numpy(
        cls,
        data: _2DArray,
        /,
        *,
        context: _LimitedContext,  # NOTE: Maybe only `Implementation`?
        schema: IntoSchema | Sequence[str] | None,
    ) -> Self:
        from narwhals.schema import Schema

        pl_schema = (
            Schema(schema).to_polars()
            if isinstance(schema, (Mapping, Schema))
            else schema
        )
        return cls.from_native(pl.from_numpy(data, pl_schema), context=context)

    def to_narwhals(self) -> DataFrame[pl.DataFrame]:
        return self._version.dataframe(self, level="full")

    def __repr__(self) -> str:  # pragma: no cover
        return "PolarsDataFrame"

    def __narwhals_dataframe__(self) -> Self:
        return self

    @overload
    def _from_native_object(self, obj: pl.Series) -> PolarsSeries: ...

    @overload
    def _from_native_object(self, obj: pl.DataFrame) -> Self: ...

    @overload
    def _from_native_object(self, obj: T) -> T: ...

    def _from_native_object(
        self, obj: pl.Series | pl.DataFrame | T
    ) -> Self | PolarsSeries | T:
        if isinstance(obj, pl.Series):
            return PolarsSeries.from_native(obj, context=self)
        if self._is_native(obj):
            return self._with_native(obj)
        # scalar
        return obj

    def __len__(self) -> int:
        return len(self.native)

    def __getattr__(self, attr: str) -> Any:
        if attr not in INHERITED_METHODS:  # pragma: no cover
            msg = f"{self.__class__.__name__} has not attribute '{attr}'."
            raise AttributeError(msg)

        def func(*args: Any, **kwargs: Any) -> Any:
            pos, kwds = extract_args_kwargs(args, kwargs)
            try:
                return self._from_native_object(getattr(self.native, attr)(*pos, **kwds))
            except pl.exceptions.ColumnNotFoundError as e:  # pragma: no cover
                msg = f"{e!s}\n\nHint: Did you mean one of these columns: {self.columns}?"
                raise ColumnNotFoundError(msg) from e
            except Exception as e:  # noqa: BLE001
                raise catch_polars_exception(e) from None

        return func

    def __array__(
        self, dtype: Any | None = None, *, copy: bool | None = None
    ) -> _2DArray:
        if self._backend_version < (0, 20, 28) and copy is not None:
            msg = "`copy` in `__array__` is only supported for 'polars>=0.20.28'"
            raise NotImplementedError(msg)
        if self._backend_version < (0, 20, 28):
            return self.native.__array__(dtype)
        return self.native.__array__(dtype)

    def to_numpy(self, dtype: Any = None, *, copy: bool | None = None) -> _2DArray:
        return self.native.to_numpy()

    @property
    def shape(self) -> tuple[int, int]:
        return self.native.shape

    def __getitem__(  # noqa: C901, PLR0912
        self,
        item: tuple[
            SingleIndexSelector | MultiIndexSelector[PolarsSeries],
            MultiColSelector[PolarsSeries],
        ],
    ) -> Any:
        rows, columns = item
        if self._backend_version > (0, 20, 30):
            rows_native = rows.native if is_compliant_series(rows) else rows
            columns_native = columns.native if is_compliant_series(columns) else columns
            selector = rows_native, columns_native
            selected = self.native.__getitem__(selector)  # type: ignore[index]
            return self._from_native_object(selected)
        else:  # pragma: no cover # noqa: RET505
            # TODO(marco): we can delete this branch after Polars==0.20.30 becomes the minimum
            # Polars version we support
            # This mostly mirrors the logic in `EagerDataFrame.__getitem__`.
            rows = list(rows) if isinstance(rows, tuple) else rows
            columns = list(columns) if isinstance(columns, tuple) else columns
            if is_numpy_array_1d(columns):
                columns = columns.tolist()

            native = self.native
            if not is_slice_none(columns):
                if isinstance(columns, Sized) and len(columns) == 0:
                    return self.select()
                if is_boolean_selector(columns):
                    native = native.select(
                        *(col for col, select in zip(native.columns, columns) if select)
                    )
                elif is_index_selector(columns):
                    if is_slice_index(columns) or is_range(columns):
                        native = native.select(
                            self.columns[slice(columns.start, columns.stop, columns.step)]
                        )
                    # NOTE: `mypy` loses track of `PolarsSeries` when `is_compliant_series` is used here
                    # `pyright` is fine
                    elif isinstance(columns, PolarsSeries):
                        native = native[:, columns.native.to_list()]
                    else:
                        native = native[:, columns]
                elif isinstance(columns, slice):
                    native = native.select(
                        self.columns[
                            slice(*convert_str_slice_to_int_slice(columns, self.columns))
                        ]
                    )
                elif is_compliant_series(columns):
                    native = native.select(columns.native.to_list())
                elif is_sequence_like(columns):
                    native = native.select(columns)
                else:
                    msg = f"Unreachable code, got unexpected type: {type(columns)}"
                    raise AssertionError(msg)

            if not is_slice_none(rows):
                if isinstance(rows, int):
                    native = native[[rows], :]
                elif isinstance(rows, (slice, range)):
                    native = native[rows, :]
                elif is_compliant_series(rows):
                    native = native[rows.native, :]
                elif is_sequence_like(rows):
                    native = native[rows, :]
                else:
                    msg = f"Unreachable code, got unexpected type: {type(rows)}"
                    raise AssertionError(msg)

            return self._with_native(native)

    def get_column(self, name: str) -> PolarsSeries:
        return PolarsSeries.from_native(self.native.get_column(name), context=self)

    def iter_columns(self) -> Iterator[PolarsSeries]:
        for series in self.native.iter_columns():
            yield PolarsSeries.from_native(series, context=self)

    def lazy(
        self,
        backend: _LazyAllowedImpl | None = None,
        *,
        session: SparkSession | None = None,
    ) -> CompliantLazyFrameAny:
        if backend is None or backend is Implementation.POLARS:
            return PolarsLazyFrame.from_native(self.native.lazy(), context=self)
        if backend is Implementation.DUCKDB:
            import duckdb  # ignore-banned-import

            from narwhals._duckdb.dataframe import DuckDBLazyFrame

            _df = self.native
            return DuckDBLazyFrame(
                duckdb.table("_df"), validate_backend_version=True, version=self._version
            )
        if backend is Implementation.DASK:
            import dask.dataframe as dd  # ignore-banned-import

            from narwhals._dask.dataframe import DaskLazyFrame

            return DaskLazyFrame(
                dd.from_pandas(self.native.to_pandas()),
                validate_backend_version=True,
                version=self._version,
            )
        if backend is Implementation.IBIS:
            import ibis  # ignore-banned-import

            from narwhals._ibis.dataframe import IbisLazyFrame

            return IbisLazyFrame(
                ibis.memtable(self.native, columns=self.columns),
                validate_backend_version=True,
                version=self._version,
            )

        if backend.is_spark_like():
            from narwhals._spark_like.dataframe import SparkLikeLazyFrame

            if session is None:
                msg = "Spark like backends require `session` to be not None."
                raise ValueError(msg)

            return SparkLikeLazyFrame._from_compliant_dataframe(
                self,  # pyright: ignore[reportArgumentType]
                session=session,
                implementation=backend,
                version=self._version,
            )

        raise AssertionError  # pragma: no cover

    @overload
    def to_dict(self, *, as_series: Literal[True]) -> dict[str, PolarsSeries]: ...

    @overload
    def to_dict(self, *, as_series: Literal[False]) -> dict[str, list[Any]]: ...

    def to_dict(
        self, *, as_series: bool
    ) -> dict[str, PolarsSeries] | dict[str, list[Any]]:
        if as_series:
            return {
                name: PolarsSeries.from_native(col, context=self)
                for name, col in self.native.to_dict().items()
            }
        return self.native.to_dict(as_series=False)

    def group_by(
        self, keys: Sequence[str] | Sequence[PolarsExpr], *, drop_null_keys: bool
    ) -> PolarsGroupBy:
        from narwhals._polars.group_by import PolarsGroupBy

        return PolarsGroupBy(self, keys, drop_null_keys=drop_null_keys)

    def drop(self, columns: Sequence[str], *, strict: bool) -> Self:
        to_drop = parse_columns_to_drop(self, columns, strict=strict)
        return self._with_native(self.native.drop(to_drop))

    @requires.backend_version((1,))
    def pivot(
        self,
        on: Sequence[str],
        *,
        index: Sequence[str] | None,
        values: Sequence[str] | None,
        aggregate_function: PivotAgg | None,
        sort_columns: bool,
        separator: str,
    ) -> Self:
        try:
            result = self.native.pivot(
                on,
                index=index,
                values=values,
                aggregate_function=aggregate_function,
                sort_columns=sort_columns,
                separator=separator,
            )
        except Exception as e:  # noqa: BLE001
            raise catch_polars_exception(e) from None
        return self._from_native_object(result)

    def to_polars(self) -> pl.DataFrame:
        return self.native

    def join(
        self,
        other: PolarsBaseFrame[pl.DataFrame],
        *,
        how: JoinStrategy,
        left_on: Sequence[str] | None,
        right_on: Sequence[str] | None,
        suffix: str,
    ) -> Self:
        try:
            return super().join(
                other=other, how=how, left_on=left_on, right_on=right_on, suffix=suffix
            )
        except Exception as e:  # noqa: BLE001
            raise catch_polars_exception(e) from None

    def top_k(
        self, k: int, *, by: str | Iterable[str], reverse: bool | Sequence[bool]
    ) -> Self:
        try:
            return super().top_k(k=k, by=by, reverse=reverse)
        except Exception as e:  # noqa: BLE001  # pragma: no cover
            raise catch_polars_exception(e) from None

    def item(self, row: int | None, column: int | str | None) -> Any:
        if (
            self._backend_version < (1, 36)
            and row is None
            and column is None
            and (shape := self.shape) != (1, 1)
        ):
            msg = (
                'can only call `.item()` without "row" or "column" values if the '
                f"DataFrame has a single element; shape={shape!r}"
            )
            raise ValueError(msg)
        return self.native.item(row=row, column=column)


class PolarsLazyFrame(PolarsBaseFrame[pl.LazyFrame]):
    sink_parquet: Method[None]

    @staticmethod
    def _is_native(obj: pl.LazyFrame | Any) -> TypeIs[pl.LazyFrame]:
        return isinstance(obj, pl.LazyFrame)

    def to_narwhals(self) -> LazyFrame[pl.LazyFrame]:
        return self._version.lazyframe(self, level="lazy")

    def __repr__(self) -> str:  # pragma: no cover
        return "PolarsLazyFrame"

    def __narwhals_lazyframe__(self) -> Self:
        return self

    def __getattr__(self, attr: str) -> Any:
        if attr not in INHERITED_METHODS:  # pragma: no cover
            msg = f"{self.__class__.__name__} has not attribute '{attr}'."
            raise AttributeError(msg)

        def func(*args: Any, **kwargs: Any) -> Any:
            pos, kwds = extract_args_kwargs(args, kwargs)
            try:
                return self._with_native(getattr(self.native, attr)(*pos, **kwds))
            except pl.exceptions.ColumnNotFoundError as e:  # pragma: no cover
                raise ColumnNotFoundError(str(e)) from e

        return func

    def _iter_columns(self) -> Iterator[PolarsSeries]:  # pragma: no cover
        yield from self.collect(Implementation.POLARS).iter_columns()

    def collect_schema(self) -> dict[str, DType]:
        try:
            return super().collect_schema()
        except Exception as e:  # noqa: BLE001
            raise catch_polars_exception(e) from None

    def collect(
        self, backend: _EagerAllowedImpl | None, **kwargs: Any
    ) -> CompliantDataFrameAny:
        try:
            result = self.native.collect(**kwargs)
        except Exception as e:  # noqa: BLE001
            raise catch_polars_exception(e) from None

        if backend is None or backend is Implementation.POLARS:
            return PolarsDataFrame.from_native(result, context=self)

        if backend is Implementation.PANDAS:
            from narwhals._pandas_like.dataframe import PandasLikeDataFrame

            return PandasLikeDataFrame(
                result.to_pandas(),
                implementation=Implementation.PANDAS,
                validate_backend_version=True,
                version=self._version,
                validate_column_names=False,
            )

        if backend is Implementation.PYARROW:
            from narwhals._arrow.dataframe import ArrowDataFrame

            return ArrowDataFrame(
                result.to_arrow(),
                validate_backend_version=True,
                version=self._version,
                validate_column_names=False,
            )

        msg = f"Unsupported `backend` value: {backend}"  # pragma: no cover
        raise ValueError(msg)  # pragma: no cover

    def group_by(
        self, keys: Sequence[str] | Sequence[PolarsExpr], *, drop_null_keys: bool
    ) -> PolarsLazyGroupBy:
        from narwhals._polars.group_by import PolarsLazyGroupBy

        return PolarsLazyGroupBy(self, keys, drop_null_keys=drop_null_keys)

    def drop(self, columns: Sequence[str], *, strict: bool) -> Self:
        if self._backend_version < (1, 0, 0):
            return self._with_native(self.native.drop(columns))
        return self._with_native(self.native.drop(columns, strict=strict))



from __future__ import annotations

import sys
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    import cudf
    import dask.dataframe as dd
    import ibis
    import modin.pandas as mpd
    import pandas as pd
    import polars as pl
    import pyarrow as pa
    from typing_extensions import TypeIs


from narwhals.dependencies import (
    IMPORT_HOOKS,
    get_cudf,
    get_dask_dataframe,
    get_ibis,
    get_modin,
    get_numpy,
    get_pandas,
    get_polars,
    get_pyarrow,
    is_into_dataframe,
    is_into_series,
    is_narwhals_dataframe,
    is_narwhals_lazyframe,
    is_narwhals_series,
    is_numpy_array,
    is_pandas_index,
)


def is_pandas_dataframe(df: Any) -> TypeIs[pd.DataFrame]:
    """Check whether `df` is a pandas DataFrame without importing pandas."""
    return ((pd := get_pandas()) is not None and isinstance(df, pd.DataFrame)) or any(
        (mod := sys.modules.get(module_name, None)) is not None
        and isinstance(df, mod.pandas.DataFrame)
        for module_name in IMPORT_HOOKS
    )


def is_pandas_series(ser: Any) -> TypeIs[pd.Series[Any]]:
    """Check whether `ser` is a pandas Series without importing pandas."""
    return ((pd := get_pandas()) is not None and isinstance(ser, pd.Series)) or any(
        (mod := sys.modules.get(module_name, None)) is not None
        and isinstance(ser, mod.pandas.Series)
        for module_name in IMPORT_HOOKS
    )


def is_modin_dataframe(df: Any) -> TypeIs[mpd.DataFrame]:
    """Check whether `df` is a modin DataFrame without importing modin."""
    return (mpd := get_modin()) is not None and isinstance(df, mpd.DataFrame)


def is_modin_series(ser: Any) -> TypeIs[mpd.Series]:
    """Check whether `ser` is a modin Series without importing modin."""
    return (mpd := get_modin()) is not None and isinstance(ser, mpd.Series)


def is_cudf_dataframe(df: Any) -> TypeIs[cudf.DataFrame]:
    """Check whether `df` is a cudf DataFrame without importing cudf."""
    return (cudf := get_cudf()) is not None and isinstance(df, cudf.DataFrame)


def is_cudf_series(ser: Any) -> TypeIs[cudf.Series[Any]]:
    """Check whether `ser` is a cudf Series without importing cudf."""
    return (cudf := get_cudf()) is not None and isinstance(ser, cudf.Series)


def is_dask_dataframe(df: Any) -> TypeIs[dd.DataFrame]:
    """Check whether `df` is a Dask DataFrame without importing Dask."""
    return (dd := get_dask_dataframe()) is not None and isinstance(df, dd.DataFrame)


def is_ibis_table(df: Any) -> TypeIs[ibis.Table]:
    """Check whether `df` is a Ibis Table without importing Ibis."""
    return (ibis := get_ibis()) is not None and isinstance(df, ibis.expr.types.Table)


def is_polars_dataframe(df: Any) -> TypeIs[pl.DataFrame]:
    """Check whether `df` is a Polars DataFrame without importing Polars."""
    return (pl := get_polars()) is not None and isinstance(df, pl.DataFrame)


def is_polars_lazyframe(df: Any) -> TypeIs[pl.LazyFrame]:
    """Check whether `df` is a Polars LazyFrame without importing Polars."""
    return (pl := get_polars()) is not None and isinstance(df, pl.LazyFrame)


def is_polars_series(ser: Any) -> TypeIs[pl.Series]:
    """Check whether `ser` is a Polars Series without importing Polars."""
    return (pl := get_polars()) is not None and isinstance(ser, pl.Series)


def is_pyarrow_chunked_array(ser: Any) -> TypeIs[pa.ChunkedArray[Any]]:
    """Check whether `ser` is a PyArrow ChunkedArray without importing PyArrow."""
    return (pa := get_pyarrow()) is not None and isinstance(ser, pa.ChunkedArray)


def is_pyarrow_table(df: Any) -> TypeIs[pa.Table]:
    """Check whether `df` is a PyArrow Table without importing PyArrow."""
    return (pa := get_pyarrow()) is not None and isinstance(df, pa.Table)


def is_pandas_like_dataframe(df: Any) -> bool:
    """Check whether `df` is a pandas-like DataFrame without doing any imports.

    By "pandas-like", we mean: pandas, Modin, cuDF.
    """
    return is_pandas_dataframe(df) or is_modin_dataframe(df) or is_cudf_dataframe(df)


def is_pandas_like_series(ser: Any) -> bool:
    """Check whether `ser` is a pandas-like Series without doing any imports.

    By "pandas-like", we mean: pandas, Modin, cuDF.
    """
    return is_pandas_series(ser) or is_modin_series(ser) or is_cudf_series(ser)


__all__ = [
    "get_cudf",
    "get_ibis",
    "get_modin",
    "get_numpy",
    "get_pandas",
    "get_polars",
    "get_pyarrow",
    "is_cudf_dataframe",
    "is_cudf_series",
    "is_dask_dataframe",
    "is_ibis_table",
    "is_into_dataframe",
    "is_into_series",
    "is_modin_dataframe",
    "is_modin_series",
    "is_narwhals_dataframe",
    "is_narwhals_lazyframe",
    "is_narwhals_series",
    "is_numpy_array",
    "is_pandas_dataframe",
    "is_pandas_index",
    "is_pandas_like_dataframe",
    "is_pandas_like_series",
    "is_pandas_series",
    "is_polars_dataframe",
    "is_polars_lazyframe",
    "is_polars_series",
    "is_pyarrow_chunked_array",
    "is_pyarrow_table",
]



from __future__ import annotations

from typing import TYPE_CHECKING

import narwhals as nw

if TYPE_CHECKING:
    from narwhals.typing import FrameT


def query(
    part_ds: FrameT,
    partsupp_ds: FrameT,
    nation_ds: FrameT,
    lineitem_ds: FrameT,
    orders_ds: FrameT,
    supplier_ds: FrameT,
) -> FrameT:
    return (
        part_ds.join(partsupp_ds, left_on="p_partkey", right_on="ps_partkey")
        .join(supplier_ds, left_on="ps_suppkey", right_on="s_suppkey")
        .join(
            lineitem_ds,
            left_on=["p_partkey", "ps_suppkey"],
            right_on=["l_partkey", "l_suppkey"],
        )
        .join(orders_ds, left_on="l_orderkey", right_on="o_orderkey")
        .join(nation_ds, left_on="s_nationkey", right_on="n_nationkey")
        .filter(nw.col("p_name").str.contains("green", literal=True))
        .select(
            nw.col("n_name").alias("nation"),
            nw.col("o_orderdate").dt.year().alias("o_year"),
            (
                nw.col("l_extendedprice") * (1 - nw.col("l_discount"))
                - nw.col("ps_supplycost") * nw.col("l_quantity")
            ).alias("amount"),
        )
        .group_by("nation", "o_year")
        .agg(nw.sum("amount").alias("sum_profit"))
        .sort(by=["nation", "o_year"], descending=[False, True])
    )



from __future__ import annotations

import re

import pytest

import narwhals as nw
from tests.utils import ConstructorEager, assert_equal_data

data = [1, 3, 2]


@pytest.mark.parametrize(("index", "expected"), [(0, 1), (1, 3)])
def test_item(constructor_eager: ConstructorEager, index: int, expected: int) -> None:
    series = nw.from_native(constructor_eager({"a": data}), eager_only=True)["a"]
    result = series.item(index)
    assert_equal_data({"a": [result]}, {"a": [expected]})
    assert_equal_data({"a": [series.head(1).item()]}, {"a": [1]})

    with pytest.raises(
        ValueError,
        match=re.escape("can only call '.item()' if the Series is of length 1,"),
    ):
        series.item(None)


def test_out_of_range(constructor_eager: ConstructorEager) -> None:
    series = nw.from_native(constructor_eager({"a": [1, 2]}), eager_only=True)["a"]
    with pytest.raises(IndexError, match="out of range"):
        series.item(2)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

import pytest

import narwhals as nw

if TYPE_CHECKING:
    from collections.abc import Mapping

    from tests.utils import ConstructorEager

pytest.importorskip("polars")
import polars as pl


@pytest.mark.filterwarnings("ignore:.*Passing a BlockManager.*:DeprecationWarning")
def test_convert_polars(constructor_eager: ConstructorEager) -> None:
    pytest.importorskip("pyarrow")
    from polars.testing import assert_frame_equal

    data: Mapping[str, Any] = {"a": [1, 3, 2], "b": [4, 4, 6], "z": [7.1, 8.0, 9.0]}
    df_raw = constructor_eager(data)
    result = nw.from_native(df_raw).to_polars()

    expected = pl.DataFrame(data)

    assert_frame_equal(result, expected)



from __future__ import annotations

from pathlib import Path
from typing import Final

from narwhals.this import ZEN

DESTINATION_PATH: Final[Path] = Path("docs") / "this.md"

content = f"""
# The Zen of Narwhals

The well famous Python easter egg `import this` will reveal The Zen of Python (PEP 20).

Narwhals took inspiration from _this_ and created its own Zen.

```py
import narwhals.this
```

```terminal
{ZEN}
```
"""

with DESTINATION_PATH.open(mode="w", encoding="utf-8", newline="\n") as destination:
    destination.write(content)



from __future__ import annotations

from typing import Generic

from narwhals.typing import SeriesT


class SeriesCatNamespace(Generic[SeriesT]):
    def __init__(self, series: SeriesT) -> None:
        self._narwhals_series = series

    def get_categories(self) -> SeriesT:
        """Get unique categories from column.

        Examples:
            >>> import pandas as pd
            >>> import narwhals as nw
            >>> s_native = pd.Series(["apple", "mango", "mango"], dtype="category")
            >>> s = nw.from_native(s_native, series_only=True)
            >>> s.cat.get_categories().to_native()
            0    apple
            1    mango
            dtype: str
        """
        return self._narwhals_series._with_compliant(
            self._narwhals_series._compliant_series.cat.get_categories()
        )



from __future__ import annotations

from datetime import date
from typing import TYPE_CHECKING, Any

import pytest

import narwhals as nw
from tests.utils import (
    CUDF_VERSION,
    DASK_VERSION,
    PANDAS_VERSION,
    POLARS_VERSION,
    PYARROW_VERSION,
    Constructor,
    assert_equal_data,
)

if TYPE_CHECKING:
    from narwhals.dtypes import DType
    from narwhals.typing import IntoDType, PythonLiteral


@pytest.mark.parametrize(
    ("dtype", "expected_lit"),
    [(None, [2, 2, 2]), (nw.String, ["2", "2", "2"]), (nw.Float32, [2.0, 2.0, 2.0])],
)
def test_lit(
    constructor: Constructor, dtype: DType | None, expected_lit: list[Any]
) -> None:
    data = {"a": [1, 3, 2], "b": [4, 4, 6], "z": [7.0, 8.0, 9.0]}
    df_raw = constructor(data)
    df = nw.from_native(df_raw).lazy()
    result = df.with_columns(nw.lit(2, dtype).alias("lit"))
    expected = {"a": [1, 3, 2], "b": [4, 4, 6], "z": [7.0, 8.0, 9.0], "lit": expected_lit}
    assert_equal_data(result, expected)


def test_lit_error(constructor: Constructor) -> None:
    pytest.importorskip("numpy")
    import numpy as np

    data = {"a": [1, 3, 2], "b": [4, 4, 6], "z": [7.0, 8.0, 9.0]}
    df_raw = constructor(data)
    df = nw.from_native(df_raw).lazy()
    with pytest.raises(
        ValueError, match="numpy arrays are not supported as literal values"
    ):
        _ = df.with_columns(nw.lit(np.array([1, 2])).alias("lit"))  # pyright: ignore[reportArgumentType]


def test_lit_out_name(constructor: Constructor) -> None:
    data = {"a": [1, 3, 2]}
    df_raw = constructor(data)
    df = nw.from_native(df_raw).lazy()
    result = df.with_columns(nw.lit(2))
    expected = {"a": [1, 3, 2], "literal": [2, 2, 2]}
    assert_equal_data(result, expected)


@pytest.mark.parametrize(
    ("col_name", "expr", "expected_result"),
    [
        ("left_lit", nw.lit(1) + nw.col("a"), [2, 4, 3]),
        ("right_lit", nw.col("a") + nw.lit(1), [2, 4, 3]),
        ("right_lit_with_abs", nw.col("a") + nw.lit(-1).abs(), [2, 4, 3]),
        ("left_lit_with_agg", nw.lit(1) + nw.col("a").mean(), [3]),
        ("right_lit_with_agg", nw.col("a").mean() - nw.lit(1), [1]),
        ("left_scalar", 1 + nw.col("a"), [2, 4, 3]),
        ("right_scalar", nw.col("a") + 1, [2, 4, 3]),
        ("left_scalar_with_agg", 1 + nw.col("a").mean(), [3]),
        ("right_scalar_with_agg", nw.col("a").mean() - 1, [1]),
        ("lit_compare", nw.col("a") == nw.lit(3), [False, True, False]),
    ],
)
def test_lit_operation_in_select(
    constructor: Constructor, col_name: str, expr: nw.Expr, expected_result: list[int]
) -> None:
    if (
        "dask" in str(constructor)
        and col_name == "right_lit_with_abs"
        and DASK_VERSION < (2025,)
    ):
        pytest.skip()

    data = {"a": [1, 3, 2]}
    df_raw = constructor(data)
    df = nw.from_native(df_raw).lazy()
    result = df.select(expr.alias(col_name))
    expected = {col_name: expected_result}
    assert_equal_data(result, expected)


@pytest.mark.parametrize(
    ("col_name", "expr", "expected_result"),
    [
        ("lit_and_scalar", (nw.lit(2) + 1), [3, 3, 3]),
        ("scalar_and_lit", (1 + nw.lit(2)), [3, 3, 3]),
    ],
)
def test_lit_operation_in_with_columns(
    constructor: Constructor, col_name: str, expr: nw.Expr, expected_result: list[int]
) -> None:
    data = {"a": [1, 3, 2]}
    df_raw = constructor(data)
    df = nw.from_native(df_raw).lazy()
    result = df.with_columns(expr.alias(col_name))
    expected = {"a": data["a"], col_name: expected_result}
    assert_equal_data(result, expected)


@pytest.mark.skipif(PANDAS_VERSION < (1, 5), reason="too old for pyarrow")
def test_date_lit(constructor: Constructor, request: pytest.FixtureRequest) -> None:
    # https://github.com/dask/dask/issues/11637
    if "dask" in str(constructor) or (
        # https://github.com/rapidsai/cudf/pull/18832
        "cudf" in str(constructor) and CUDF_VERSION >= (25, 8, 0)
    ):
        request.applymarker(pytest.mark.xfail)
    if "pandas" in str(constructor):
        pytest.importorskip("pyarrow")
    df = nw.from_native(constructor({"a": [1]}))
    result = df.with_columns(nw.lit(date(2020, 1, 1), dtype=nw.Date)).collect_schema()
    if df.implementation.is_cudf():
        # cudf has no date dtype
        assert result == {"a": nw.Int64, "literal": nw.Datetime}
    else:
        assert result == {"a": nw.Int64, "literal": nw.Date}


def test_pyarrow_lit_string() -> None:
    pytest.importorskip("pyarrow")
    import pyarrow as pa

    df = nw.from_native(pa.table({"a": [1, 2, 3]}))
    result = df.select(nw.lit("foo")).to_native().schema.field("literal")
    assert pa.types.is_string(result.type)
    result = df.select(nw.lit("foo", dtype=nw.String)).to_native().schema.field("literal")
    assert pa.types.is_string(result.type)


@pytest.mark.parametrize(
    ("value", "dtype"),
    [
        # Empty nested structures
        ((), nw.List(nw.Int32())),
        ([], nw.List(nw.Int32())),
        ({}, nw.Struct({})),
        # Nested structures with different size from dataframe
        (("foo", "bar"), None),
        (["orca", "narwhal"], None),
        ({"field_1": 42}, None),
        # Nested structures with same size as dataframe
        (("foo", "bar", "baz"), None),
        (["orca", "narwhal", "penguin"], None),
        (
            {"field_1": 42, "field_2": 1.2, "field_3": True},
            nw.Struct(
                {"field_1": nw.Int32(), "field_2": nw.Float64(), "field_3": nw.Boolean()}
            ),
        ),
    ],
)
def test_nested_structures(
    request: pytest.FixtureRequest,
    constructor: Constructor,
    value: PythonLiteral,
    dtype: IntoDType | None,
) -> None:
    is_empty_dict = isinstance(value, dict) and len(value) == 0
    non_pyspark_sql_like = ("duckdb", "sqlframe", "ibis")
    is_non_pyspark_sql_like = any(x in str(constructor) for x in non_pyspark_sql_like)
    if is_non_pyspark_sql_like and is_empty_dict:
        reason = "Cannot create an empty struct type for backend"
        request.applymarker(pytest.mark.xfail(reason=reason, raises=NotImplementedError))

    # TODO(FBruzzesi): Check cudf
    if any(x in str(constructor) for x in ("cudf", "dask")):
        reason = "Nested structures are not support for backend"
        request.applymarker(pytest.mark.xfail(reason=reason, raises=NotImplementedError))

    if any(x in str(constructor) for x in ("pandas", "modin")) and (
        PYARROW_VERSION == (0, 0, 0) or PANDAS_VERSION < (2, 0)
    ):  # pragma: no cover
        reason = "Requires pyarrow and pandas 2.0+"
        pytest.skip(reason=reason)

    if (
        "polars" in str(constructor)
        and isinstance(value, dict)
        and POLARS_VERSION < (1, 10, 0)
    ):  # pragma: no cover
        reason = "polars<1.10 does not support dict to struct in lit"
        pytest.skip(reason=reason)

    size = 3
    data = {"a": list(range(size))}
    expr = nw.lit(value, dtype=dtype).alias("nested")

    value_ = list(value) if isinstance(value, tuple) else value
    expected_nested = {"nested": [value_] * size}

    frame = nw.from_native(constructor(data))

    result_with_cols = frame.with_columns(expr)
    assert_equal_data(result_with_cols, {**data, **expected_nested})

    result_select = frame.select(expr, nw.col("a"))
    assert_equal_data(result_select, {**expected_nested, **data})


@pytest.mark.parametrize("value", [[], (), {}])
def test_raise_empty_nested_structures(value: PythonLiteral) -> None:
    msg = "Cannot infer dtype for empty nested structure. Please provide an explicit dtype parameter."
    with pytest.raises(ValueError, match=msg):
        nw.lit(value=value)


@pytest.mark.parametrize(
    "value",
    [
        # List containing nested structures
        [[1, 2], [3, 4]],
        [(1, 2), (3, 4)],
        [{"a": 1}, {"a": 2}],
        # Tuple containing nested structures
        ([1, 2], [3, 4]),
        ((1, 2), (3, 4)),
        ({"a": 1}, {"a": 2}),
        # Dict containing nested structures
        {"a": [1, 2], "b": [3, 4]},
        {"a": (1, 2), "b": (3, 4)},
        {"a": {"x": 1}, "b": {"y": 2}},
    ],
)
def test_raise_nested_structures_with_nested_values(value: Any) -> None:
    msg = "Nested structures with nested values are not supported."
    with pytest.raises(NotImplementedError, match=msg):
        nw.lit(value=value)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

import pytest

import narwhals as nw
from tests.utils import is_pd_na

if TYPE_CHECKING:
    from tests.utils import ConstructorEager

data = {"a": [1, 3, 2], "b": [4, 4, 6], "z": [7.0, 8.0, 9.0]}
data_na = {"a": [None, 3, 2], "b": [4, 4, 6], "z": [7.0, None, 9]}


@pytest.mark.parametrize(
    ("named", "expected"),
    [
        (False, [(1, 4, 7.0, 5), (3, 4, 8.0, 6), (2, 6, 9.0, 7)]),
        (
            True,
            [
                {"a": 1, "_b": 4, "z": 7.0, "1": 5},
                {"a": 3, "_b": 4, "z": 8.0, "1": 6},
                {"a": 2, "_b": 6, "z": 9.0, "1": 7},
            ],
        ),
    ],
)
def test_iter_rows(
    request: Any,
    constructor_eager: ConstructorEager,
    named: bool,  # noqa: FBT001
    expected: list[tuple[Any, ...]] | list[dict[str, Any]],
) -> None:
    if "cudf" in str(constructor_eager):
        request.applymarker(pytest.mark.xfail)

    data = {"a": [1, 3, 2], "_b": [4, 4, 6], "z": [7.0, 8.0, 9.0], "1": [5, 6, 7]}
    df = nw.from_native(constructor_eager(data), eager_only=True)
    result = list(df.iter_rows(named=named))
    assert result == expected


@pytest.mark.filterwarnings(
    "ignore:.*all arguments of to_dict except for the argument:FutureWarning"
)
@pytest.mark.parametrize(
    ("named", "expected"),
    [
        (False, [(1, 4, 7.0), (3, 4, 8.0), (2, 6, 9.0)]),
        (
            True,
            [
                {"a": 1, "b": 4, "z": 7.0},
                {"a": 3, "b": 4, "z": 8.0},
                {"a": 2, "b": 6, "z": 9.0},
            ],
        ),
    ],
)
def test_rows(
    constructor_eager: ConstructorEager,
    named: bool,  # noqa: FBT001
    expected: list[tuple[Any, ...]] | list[dict[str, Any]],
) -> None:
    df = nw.from_native(constructor_eager(data), eager_only=True)
    result = df.rows(named=named)
    assert result == expected


@pytest.mark.filterwarnings(
    r"ignore:.*Starting with pandas version 3\.0 all arguments of to_dict"
)
@pytest.mark.parametrize(
    ("named", "expected"),
    [
        (False, [(1, 4, 7.0, 5), (3, 4, 8.0, 6), (2, 6, 9.0, 7)]),
        (
            True,
            [
                {"a": 1, "_b": 4, "z": 7.0, "1": 5},
                {"a": 3, "_b": 4, "z": 8.0, "1": 6},
                {"a": 2, "_b": 6, "z": 9.0, "1": 7},
            ],
        ),
    ],
)
def test_rows_eager(
    constructor_eager: Any,
    named: bool,  # noqa: FBT001
    expected: list[tuple[Any, ...]] | list[dict[str, Any]],
) -> None:
    # posit-dev/py-shiny relies on `.rows(named=False)` to return unnamed rows
    data = {"a": [1, 3, 2], "_b": [4, 4, 6], "z": [7.0, 8.0, 9.0], "1": [5, 6, 7]}
    df = nw.from_native(constructor_eager(data), eager_only=True)
    result = df.rows(named=named)
    assert result == expected


def test_rows_with_nulls_unnamed(
    constructor_eager: ConstructorEager, request: pytest.FixtureRequest
) -> None:
    if "cudf" in str(constructor_eager):
        # cudf intentionally doesn't support itertuples / iter_rows
        request.applymarker(pytest.mark.xfail)
    df = nw.from_native(constructor_eager(data_na), eager_only=True)
    result = list(df.iter_rows(named=False))
    expected = [(None, 4, 7.0), (3, 4, None), (2, 6, 9.0)]
    for i, row in enumerate(expected):
        for j, value in enumerate(row):
            value_in_result = result[i][j]
            if value is None:
                assert value_in_result is None or is_pd_na(
                    value_in_result
                )  # because float('nan') != float('nan')
            else:
                assert value_in_result == value


def test_rows_with_nulls_named(
    constructor_eager: ConstructorEager, request: pytest.FixtureRequest
) -> None:
    if "cudf" in str(constructor_eager):
        # cudf intentionally doesn't support itertuples / iter_rows
        request.applymarker(pytest.mark.xfail)
    df = nw.from_native(constructor_eager(data_na), eager_only=True)
    result = list(df.iter_rows(named=True))
    expected: list[dict[str, Any]] = [
        {"a": None, "b": 4, "z": 7.0},
        {"a": 3, "b": 4, "z": None},
        {"a": 2, "b": 6, "z": 9.0},
    ]
    for i, row in enumerate(expected):
        for col, value in row.items():
            value_in_result = result[i][col]
            if value is None:
                assert value_in_result is None or is_pd_na(
                    value_in_result
                )  # because float('nan') != float('nan')
            else:
                assert value_in_result == value



from __future__ import annotations

import pytest

import narwhals as nw
from tests.utils import ConstructorEager, assert_equal_data

data = {"a": list(range(10))}


@pytest.mark.parametrize("n", [1, 2, 3])
@pytest.mark.parametrize("offset", [1, 2, 3])
def test_gather_every_series(
    constructor_eager: ConstructorEager, n: int, offset: int
) -> None:
    series = nw.from_native(constructor_eager(data), eager_only=True)["a"]

    result = series.gather_every(n=n, offset=offset)
    expected = data["a"][offset::n]

    assert_equal_data({"a": result}, {"a": expected})



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

import narwhals as nw
from tests.utils import PANDAS_VERSION, POLARS_VERSION, assert_equal_data

if TYPE_CHECKING:
    from typing import Any

    from tests.utils import Constructor, ConstructorEager


data = {"a": [[3, 2, 2, 4, -10, None, None], [-1], None, [None, None, None], []]}
expected_desc_nulls_last = [
    [4, 3, 2, 2, -10, None, None],
    [-1],
    None,
    [None, None, None],
    [],
]
expected_desc_nulls_first = [
    [None, None, 4, 3, 2, 2, -10],
    [-1],
    None,
    [None, None, None],
    [],
]
expected_asc_nulls_last = [
    [-10, 2, 2, 3, 4, None, None],
    [-1],
    None,
    [None, None, None],
    [],
]
expected_asc_nulls_first = [
    [None, None, -10, 2, 2, 3, 4],
    [-1],
    None,
    [None, None, None],
    [],
]


@pytest.mark.parametrize(
    ("descending", "nulls_last", "expected"),
    [
        (True, True, expected_desc_nulls_last),
        (True, False, expected_desc_nulls_first),
        (False, True, expected_asc_nulls_last),
        (False, False, expected_asc_nulls_first),
    ],
)
def test_sort_expr_args(
    request: pytest.FixtureRequest,
    constructor: Constructor,
    descending: bool,  # noqa: FBT001
    nulls_last: bool,  # noqa: FBT001
    expected: list[Any],
) -> None:
    if any(backend in str(constructor) for backend in ("dask", "cudf")):
        request.applymarker(pytest.mark.xfail)
    if "ibis" in str(constructor) and descending:
        # https://github.com/ibis-project/ibis/issues/11735
        request.applymarker(pytest.mark.xfail)
    if "polars" in str(constructor) and POLARS_VERSION < (0, 20, 5):
        pytest.skip()
    if "pandas" in str(constructor):
        if PANDAS_VERSION < (2, 2):
            pytest.skip()
        pytest.importorskip("pyarrow")
    result = nw.from_native(constructor(data)).select(
        nw.col("a")
        .cast(nw.List(nw.Int32()))
        .list.sort(descending=descending, nulls_last=nulls_last)
    )
    assert_equal_data(result, {"a": expected})


@pytest.mark.parametrize(
    ("descending", "nulls_last", "expected"),
    [
        (True, True, expected_desc_nulls_last),
        (True, False, expected_desc_nulls_first),
        (False, True, expected_asc_nulls_last),
        (False, False, expected_asc_nulls_first),
    ],
)
def test_sort_series_args(
    request: pytest.FixtureRequest,
    constructor_eager: ConstructorEager,
    descending: bool,  # noqa: FBT001
    nulls_last: bool,  # noqa: FBT001
    expected: list[Any],
) -> None:
    if any(backend in str(constructor_eager) for backend in ("dask", "cudf")):
        request.applymarker(pytest.mark.xfail)
    if "polars" in str(constructor_eager) and POLARS_VERSION < (0, 20, 5):
        pytest.skip()
    if "pandas" in str(constructor_eager):
        if PANDAS_VERSION < (2, 2):
            pytest.skip()
        pytest.importorskip("pyarrow")
    df = nw.from_native(constructor_eager(data), eager_only=True)
    result = (
        df["a"]
        .cast(nw.List(nw.Int32()))
        .list.sort(descending=descending, nulls_last=nulls_last)
    )
    assert_equal_data({"a": result}, {"a": expected})



from __future__ import annotations

import pytest

import narwhals as nw
from tests.utils import Constructor, assert_equal_data

data = {"weight": ["89kg", "83", "79kg", "68kg", "78kg", "73", "86kg"]}

expected = {"weight_kg": [89, 38, 79, 68, 78, 33, 86]}


def test_parse_weight(request: pytest.FixtureRequest, constructor: Constructor) -> None:
    if any(x in str(constructor) for x in ("pyspark", "ibis", "duckdb")):
        request.applymarker(pytest.mark.xfail)
    df = nw.from_native(constructor(data))

    col = nw.col("weight")
    is_kg = col.str.ends_with("kg")
    without_unit = col.str.replace("kg", "").cast(nw.UInt8)
    pounds_to_kg = (without_unit * 0.453592).round(0).cast(nw.UInt8)

    result = df.select(
        nw.when(is_kg).then(without_unit).otherwise(pounds_to_kg).alias("weight_kg")
    )

    assert_equal_data(result, expected)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Generic

from narwhals._compliant import LazyExprNamespace
from narwhals._compliant.any_namespace import DateTimeNamespace
from narwhals._sql.typing import SQLExprT

if TYPE_CHECKING:
    from narwhals._compliant.expr import NativeExpr


class SQLExprDateTimeNamesSpace(
    LazyExprNamespace[SQLExprT], DateTimeNamespace[SQLExprT], Generic[SQLExprT]
):
    def _function(self, name: str, *args: Any) -> NativeExpr:
        return self.compliant._function(name, *args)  # type: ignore[no-any-return]

    def year(self) -> SQLExprT:
        return self.compliant._with_elementwise(lambda expr: self._function("year", expr))

    def month(self) -> SQLExprT:
        return self.compliant._with_elementwise(
            lambda expr: self._function("month", expr)
        )

    def day(self) -> SQLExprT:
        return self.compliant._with_elementwise(lambda expr: self._function("day", expr))

    def hour(self) -> SQLExprT:
        return self.compliant._with_elementwise(lambda expr: self._function("hour", expr))

    def minute(self) -> SQLExprT:
        return self.compliant._with_elementwise(
            lambda expr: self._function("minute", expr)
        )

    def second(self) -> SQLExprT:
        return self.compliant._with_elementwise(
            lambda expr: self._function("second", expr)
        )

    def ordinal_day(self) -> SQLExprT:
        return self.compliant._with_elementwise(
            lambda expr: self._function("dayofyear", expr)
        )

    def date(self) -> SQLExprT:
        return self.compliant._with_elementwise(
            lambda expr: self._function("to_date", expr)
        )



from __future__ import annotations

from typing import TYPE_CHECKING, Callable

import pytest

import narwhals as nw

if TYPE_CHECKING:
    import pandas as pd


def test_ops_preserve_column_index_name(
    constructor: Callable[..., pd.DataFrame], request: pytest.FixtureRequest
) -> None:
    if not any(x in str(constructor) for x in ("pandas", "modin", "cudf", "dask")):
        pytest.skip(
            reason="Dataframe columns is a list and do not have a `name` like a pandas Index does"
        )
    if "dask" in str(constructor):
        # https://github.com/dask/dask/issues/11874
        request.applymarker(pytest.mark.xfail)

    data = {"a": [1, 3, 2], "b": [4, 4, 6], "z": [7.0, 8.0, 9.0]}
    df_native = constructor(data)
    df_native.columns.name = "foo"

    df = nw.from_native(df_native)

    result = df.with_columns(b=nw.col("a") + 1, c=nw.col("a") * 2).select("c", "b")

    assert result.to_native().columns.name == "foo"
    assert result.lazy().collect(backend="pandas").to_native().columns.name == "foo"



"""Tricks to generate nan's and null's for pandas with nullable backends.

* Square rooting a negative number will generate a NaN
* Replacing a value with None once the dtype is nullable will generate <NA>'s
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

import pytest

import narwhals as nw
from narwhals.exceptions import ComputeError, InvalidOperationError
from tests.conftest import (
    dask_lazy_p1_constructor,
    dask_lazy_p2_constructor,
    modin_constructor,
    pandas_constructor,
)
from tests.utils import (
    PANDAS_VERSION,
    PYARROW_VERSION,
    Constructor,
    ConstructorEager,
    assert_equal_data,
)

if TYPE_CHECKING:
    from narwhals.typing import NumericLiteral

NON_NULLABLE_CONSTRUCTORS = (
    pandas_constructor,
    dask_lazy_p1_constructor,
    dask_lazy_p2_constructor,
    modin_constructor,
)
NULL_PLACEHOLDER, NAN_PLACEHOLDER = 9999.0, -1.0
INF_POS, INF_NEG = float("inf"), float("-inf")

data: dict[str, Any] = {
    "x": [1.001, NULL_PLACEHOLDER, NAN_PLACEHOLDER, INF_POS, INF_NEG, INF_POS],
    "y": [1.005, NULL_PLACEHOLDER, NAN_PLACEHOLDER, INF_POS, 3.0, INF_NEG],
    "non_numeric": list("number"),
    "idx": list(range(6)),
}


# Exceptions
def test_is_close_series_raise_non_numeric(constructor_eager: ConstructorEager) -> None:
    df = nw.from_native(constructor_eager(data), eager_only=True)
    x, y = df["non_numeric"], df["y"]

    msg = "`is_close` operation not supported for dtype"
    with pytest.raises(InvalidOperationError, match=msg):
        x.is_close(y)


@pytest.mark.parametrize("rel_tol", [1e-09, 999])
def test_is_close_raise_negative_abs_tol(
    constructor_eager: ConstructorEager, rel_tol: float
) -> None:
    df = nw.from_native(constructor_eager(data), eager_only=True)
    x, y = df["x"], df["y"]

    abs_tol = -2
    msg = rf"`abs_tol` must be non-negative but got {abs_tol}"
    with pytest.raises(ComputeError, match=msg):
        x.is_close(y, abs_tol=abs_tol, rel_tol=rel_tol)

    with pytest.raises(ComputeError, match=msg):
        df.select(nw.col("x").is_close(nw.col("y"), abs_tol=abs_tol, rel_tol=rel_tol))


@pytest.mark.parametrize("rel_tol", [-0.0001, 1.0, 1.1])
def test_is_close_raise_invalid_rel_tol(
    constructor_eager: ConstructorEager, rel_tol: float
) -> None:
    df = nw.from_native(constructor_eager(data), eager_only=True)
    x, y = df["x"], df["y"]

    msg = rf"`rel_tol` must be in the range \[0, 1\) but got {rel_tol}"
    with pytest.raises(ComputeError, match=msg):
        x.is_close(y, rel_tol=rel_tol)

    with pytest.raises(ComputeError, match=msg):
        df.select(nw.col("x").is_close(nw.col("y"), rel_tol=rel_tol))


cases_columnar = pytest.mark.parametrize(
    ("abs_tol", "rel_tol", "nans_equal", "expected"),
    [
        (0.1, 0.0, False, [True, None, False, True, False, False]),
        (0.0001, 0.0, True, [False, None, True, True, False, False]),
        (0.0, 0.1, False, [True, None, False, True, False, False]),
        (0.0, 0.001, True, [False, None, True, True, False, False]),
    ],
)
cases_scalar = pytest.mark.parametrize(
    ("other", "abs_tol", "rel_tol", "nans_equal", "expected"),
    [
        (1.0, 0.1, 0.0, False, [True, None, False, False, False, False]),
        (1.0, 0.0001, 0.0, True, [False, None, False, False, False, False]),
        (2.9, 0.0, 0.1, False, [False, None, False, False, True, False]),
        (2.9, 0.0, 0.001, True, [False, None, False, False, False, False]),
    ],
)


# Series
@cases_columnar
def test_is_close_series_with_series(
    constructor_eager: ConstructorEager,
    abs_tol: float,
    rel_tol: float,
    *,
    nans_equal: bool,
    expected: list[Any],
) -> None:
    df = nw.from_native(constructor_eager(data), eager_only=True)
    x, y = df["x"], df["y"]
    nulls = nw.new_series(
        "nulls", [None] * len(x), nw.Float64(), backend=df.implementation
    )
    x = x.zip_with(x != NAN_PLACEHOLDER, x**0.5).zip_with(x != NULL_PLACEHOLDER, nulls)
    y = y.zip_with(y != NAN_PLACEHOLDER, y**0.5).zip_with(y != NULL_PLACEHOLDER, nulls)
    result = x.is_close(y, abs_tol=abs_tol, rel_tol=rel_tol, nans_equal=nans_equal)

    if constructor_eager in NON_NULLABLE_CONSTRUCTORS:
        expected = [v if v is not None else nans_equal for v in expected]
    elif "pandas" in str(constructor_eager) and PANDAS_VERSION >= (3,):
        expected = [
            v if data["y"][i] not in {NULL_PLACEHOLDER, NAN_PLACEHOLDER} else None
            for i, v in enumerate(expected)
        ]
    assert_equal_data({"result": result}, {"result": expected})


@cases_scalar
def test_is_close_series_with_scalar(
    constructor_eager: ConstructorEager,
    other: NumericLiteral,
    abs_tol: float,
    rel_tol: float,
    *,
    nans_equal: bool,
    expected: list[Any],
) -> None:
    df = nw.from_native(constructor_eager(data), eager_only=True)
    y = df["y"]
    nulls = nw.new_series(
        "nulls", [None] * len(y), nw.Float64(), backend=df.implementation
    )
    y = y.zip_with(y != NAN_PLACEHOLDER, y**0.5).zip_with(y != NULL_PLACEHOLDER, nulls)
    result = y.is_close(other, abs_tol=abs_tol, rel_tol=rel_tol, nans_equal=nans_equal)

    if constructor_eager in NON_NULLABLE_CONSTRUCTORS:
        expected = [v if v is not None else False for v in expected]
    elif "pandas" in str(constructor_eager) and PANDAS_VERSION >= (3,):
        expected = [
            v if data["y"][i] not in {NULL_PLACEHOLDER, NAN_PLACEHOLDER} else None
            for i, v in enumerate(expected)
        ]
    assert_equal_data({"result": result}, {"result": expected})


# Expr
@cases_columnar
def test_is_close_expr_with_expr(
    request: pytest.FixtureRequest,
    constructor: Constructor,
    abs_tol: float,
    rel_tol: float,
    *,
    nans_equal: bool,
    expected: list[Any],
) -> None:
    if "sqlframe" in str(constructor):
        # TODO(FBruzzesi): Figure out a MRE and report upstream
        reason = (
            "duckdb.duckdb.ParserException: Parser Error: syntax error at or near '='"
        )
        request.applymarker(pytest.mark.xfail(reason=reason))

    x, y = nw.col("x"), nw.col("y")
    result = (
        nw.from_native(constructor(data))
        .with_columns(
            x=nw.when(x != NAN_PLACEHOLDER).then(x).otherwise(x**0.5),
            y=nw.when(y != NAN_PLACEHOLDER).then(y).otherwise(y**0.5),
        )
        .with_columns(
            x=nw.when(x != NULL_PLACEHOLDER).then(x),
            y=nw.when(y != NULL_PLACEHOLDER).then(y),
        )
        .select(
            "idx",
            result=x.is_close(y, abs_tol=abs_tol, rel_tol=rel_tol, nans_equal=nans_equal),
        )
        .sort("idx")
    )
    if constructor in NON_NULLABLE_CONSTRUCTORS:
        expected = [v if v is not None else nans_equal for v in expected]
    elif "pandas" in str(constructor) and PANDAS_VERSION >= (3,):
        expected = [
            v if data["y"][i] not in {NULL_PLACEHOLDER, NAN_PLACEHOLDER} else None
            for i, v in enumerate(expected)
        ]
    assert_equal_data(result, {"idx": data["idx"], "result": expected})


@cases_scalar
def test_is_close_expr_with_scalar(
    request: pytest.FixtureRequest,
    constructor: Constructor,
    other: NumericLiteral,
    abs_tol: float,
    rel_tol: float,
    *,
    nans_equal: bool,
    expected: list[Any],
) -> None:
    if "sqlframe" in str(constructor):
        # TODO(FBruzzesi): Figure out a MRE and report upstream
        reason = (
            "duckdb.duckdb.ParserException: Parser Error: syntax error at or near '='"
        )
        request.applymarker(pytest.mark.xfail(reason=reason))

    y = nw.col("y")
    result = (
        nw.from_native(constructor(data))
        .with_columns(y=nw.when(y != NAN_PLACEHOLDER).then(y).otherwise(y**0.5))
        .with_columns(y=nw.when(y != NULL_PLACEHOLDER).then(y))
        .select(
            "idx",
            result=y.is_close(
                other, abs_tol=abs_tol, rel_tol=rel_tol, nans_equal=nans_equal
            ),
        )
        .sort("idx")
    )
    if constructor in NON_NULLABLE_CONSTRUCTORS:
        expected = [v if v is not None else False for v in expected]
    elif "pandas" in str(constructor) and PANDAS_VERSION >= (3,):
        expected = [
            v if data["y"][i] not in {NULL_PLACEHOLDER, NAN_PLACEHOLDER} else None
            for i, v in enumerate(expected)
        ]
    assert_equal_data(result, {"idx": data["idx"], "result": expected})


def test_is_close_pandas_unnamed() -> None:
    pytest.importorskip("pandas")
    import pandas as pd

    ser = nw.from_native(pd.Series([1.1, 1.2]), series_only=True)
    res = ser.is_close(ser)
    assert res.name is None
    ser = nw.from_native(pd.Series([1.1, 1.2], name="ab"), series_only=True)
    res = ser.is_close(ser)
    assert res.name == "ab"


def test_issue_3474_series_decimal(constructor_eager: ConstructorEager) -> None:
    frame = nw.from_native(constructor_eager({"a": [0, 1, 2]}))

    if frame.implementation.is_pandas_like() and (
        PYARROW_VERSION == (0, 0, 0) or PANDAS_VERSION < (2, 2)
    ):
        pytest.skip(reason="pyarrow is required to convert to decimal dtype")

    frame = frame.with_columns(nw.col("a").cast(nw.Decimal()))
    assert frame["a"].is_close(frame["a"]).all()


def test_issue_3474_expr_decimal(
    constructor: Constructor, request: pytest.FixtureRequest
) -> None:
    if any(x in str(constructor) for x in ("dask", "sqlframe")):
        # TODO(FBruzzesi): Figure out a MRE and report upstream
        reason = (
            "SQLFrame: duckdb.duckdb.ParserException: Parser Error: syntax error at or near '='\n"
            "Dask: Converting to Decimal dtype is not supported."
        )
        request.applymarker(pytest.mark.xfail(reason=reason))

    frame = nw.from_native(constructor({"a": [0, 1, 2]}))

    if frame.implementation.is_pandas_like() and (
        PYARROW_VERSION == (0, 0, 0) or PANDAS_VERSION < (2, 2)
    ):
        pytest.skip(reason="pyarrow is required to convert to decimal dtype")

    frame = frame.lazy().with_columns(nw.col("a").cast(nw.Decimal()))
    assert frame.select((nw.col("a").is_close(nw.col("a"))).all()).collect().item()



from __future__ import annotations

from typing import cast

import pytest

import narwhals as nw
from tests.utils import PANDAS_VERSION, Constructor, ConstructorEager, assert_equal_data


def test_get_field_expr(request: pytest.FixtureRequest, constructor: Constructor) -> None:
    pytest.importorskip("pyarrow")
    import pyarrow as pa

    if any(backend in str(constructor) for backend in ("dask", "modin")):
        request.applymarker(pytest.mark.xfail)
    if "pandas" in str(constructor) and PANDAS_VERSION < (2, 2, 0):
        pytest.skip()
    data = {"user": [{"id": "0", "name": "john"}, {"id": "1", "name": "jane"}]}

    df_native = constructor(data)

    if "pandas" in str(constructor):
        import pandas as pd

        df_native = cast("pd.DataFrame", df_native).assign(
            user=pd.Series(
                data["user"],
                dtype=pd.ArrowDtype(
                    pa.struct([("id", pa.string()), ("name", pa.string())])
                ),
            )
        )

    df = nw.from_native(df_native)

    result = nw.from_native(df).select(
        nw.col("user").struct.field("id"), nw.col("user").struct.field("name")
    )
    expected = {"id": ["0", "1"], "name": ["john", "jane"]}
    assert_equal_data(result, expected)
    result = nw.from_native(df).select(nw.col("user").struct.field("id").name.keep())
    expected = {"user": ["0", "1"]}
    assert_equal_data(result, expected)


def test_get_field_series(
    request: pytest.FixtureRequest, constructor_eager: ConstructorEager
) -> None:
    pytest.importorskip("pyarrow")
    import pyarrow as pa

    if any(backend in str(constructor_eager) for backend in ("modin",)):
        request.applymarker(pytest.mark.xfail)
    if "pandas" in str(constructor_eager) and PANDAS_VERSION < (2, 2, 0):
        pytest.skip()
    data = {"user": [{"id": "0", "name": "john"}, {"id": "1", "name": "jane"}]}
    expected = {"id": ["0", "1"], "name": ["john", "jane"]}

    _expected = expected.copy()
    df_native = constructor_eager(data)

    if "pandas" in str(constructor_eager):
        import pandas as pd

        df_native = cast("pd.DataFrame", df_native).assign(
            user=pd.Series(
                data["user"],
                dtype=pd.ArrowDtype(
                    pa.struct([("id", pa.string()), ("name", pa.string())])
                ),
            )
        )

    df = nw.from_native(df_native, eager_only=True)

    result = nw.from_native(df).select(
        df["user"].struct.field("id"), df["user"].struct.field("name")
    )
    expected = {"id": ["0", "1"], "name": ["john", "jane"]}
    assert_equal_data(result, _expected)


def test_pandas_object_series() -> None:
    pytest.importorskip("pandas")
    import pandas as pd

    s_native = pd.Series(data=[{"id": "0", "name": "john"}, {"id": "1", "name": "jane"}])
    s = nw.from_native(s_native, series_only=True)

    with pytest.raises(TypeError):
        s.struct.field("name")



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

pytest.importorskip("pandas")
import pandas as pd
from pandas.testing import assert_series_equal

import narwhals as nw
from tests.utils import PANDAS_VERSION

if TYPE_CHECKING:
    from tests.utils import ConstructorEager

data = [1, 3, 2]


@pytest.mark.skipif(PANDAS_VERSION < (2, 0, 0), reason="too old for pyarrow")
def test_convert(
    request: pytest.FixtureRequest, constructor_eager: ConstructorEager
) -> None:
    pytest.importorskip("pyarrow")

    if any(
        cname in str(constructor_eager)
        for cname in ("pandas_nullable", "pandas_pyarrow", "modin_pyarrow")
    ):
        request.applymarker(pytest.mark.xfail)

    series = nw.from_native(constructor_eager({"a": data}), eager_only=True)["a"].alias(
        "a"
    )

    result = series.to_pandas()
    assert_series_equal(result, pd.Series([1, 3, 2], name="a"))



from __future__ import annotations

from typing import TYPE_CHECKING, Any, NoReturn

from narwhals._ibis.utils import native_to_narwhals_dtype
from narwhals.dependencies import get_ibis

if TYPE_CHECKING:
    from types import ModuleType

    from typing_extensions import Self

    from narwhals._utils import Version
    from narwhals.dtypes import DType


class IbisInterchangeSeries:
    def __init__(self, df: Any, version: Version) -> None:
        self._native_series = df
        self._version = version

    def __narwhals_series__(self) -> Self:
        return self

    def __native_namespace__(self) -> ModuleType:
        return get_ibis()

    @property
    def dtype(self) -> DType:
        return native_to_narwhals_dtype(
            self._native_series.schema().types[0], self._version
        )

    def __getattr__(self, attr: str) -> NoReturn:
        msg = (
            f"Attribute {attr} is not supported for interchange-level dataframes.\n\n"
            "If you would like to see this kind of object better supported in "
            "Narwhals, please open a feature request "
            "at https://github.com/narwhals-dev/narwhals/issues."
        )
        raise NotImplementedError(msg)



from __future__ import annotations

import pytest

import narwhals as nw
from tests.utils import Constructor, assert_equal_data


def test_format(constructor: Constructor) -> None:
    df = nw.from_native(
        constructor(
            {
                "name": ["bob", "alice", "dodo"],
                "surname": ["builder", "wonderlander", "extinct"],
            }
        )
    )
    result = df.select(fmt=nw.format("hello {} {} wassup", "name", nw.col("surname")))
    expected = {
        "fmt": [
            "hello bob builder wassup",
            "hello alice wonderlander wassup",
            "hello dodo extinct wassup",
        ]
    }
    assert_equal_data(result, expected)
    result = df.select(fmt=nw.format("{} {} wassup", "name", nw.col("surname")))
    expected = {
        "fmt": ["bob builder wassup", "alice wonderlander wassup", "dodo extinct wassup"]
    }
    assert_equal_data(result, expected)


def test_format_invalid() -> None:
    pytest.importorskip("pandas")
    import pandas as pd

    df = nw.from_native(
        pd.DataFrame(
            {
                "name": ["bob", "alice", "dodo"],
                "surname": ["builder", "wonderlander", "extinct"],
            }
        )
    )
    with pytest.raises(ValueError, match="Expected 2 arguments, got 1"):
        df.select(fmt=nw.format("hello {} {} wassup", "name"))



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING

import narwhals as nw

if TYPE_CHECKING:
    from narwhals.typing import FrameT


def query(lineitem: FrameT) -> FrameT:
    var_1 = datetime(1998, 9, 2)
    return (
        lineitem.filter(nw.col("l_shipdate") <= var_1)
        .with_columns(
            disc_price=nw.col("l_extendedprice") * (1 - nw.col("l_discount")),
            charge=(
                nw.col("l_extendedprice")
                * (1.0 - nw.col("l_discount"))
                * (1.0 + nw.col("l_tax"))
            ),
        )
        .group_by("l_returnflag", "l_linestatus")
        .agg(
            nw.sum("l_quantity").alias("sum_qty"),
            nw.sum("l_extendedprice").alias("sum_base_price"),
            nw.sum("disc_price").alias("sum_disc_price"),
            nw.sum("charge").alias("sum_charge"),
            nw.mean("l_quantity").alias("avg_qty"),
            nw.mean("l_extendedprice").alias("avg_price"),
            nw.mean("l_discount").alias("avg_disc"),
            nw.len().alias("count_order"),
        )
        .sort("l_returnflag", "l_linestatus")
    )



from __future__ import annotations

from typing import TYPE_CHECKING

from narwhals._compliant import LazyExprNamespace
from narwhals._compliant.any_namespace import DateTimeNamespace
from narwhals._constants import MS_PER_SECOND, NS_PER_SECOND, US_PER_SECOND
from narwhals._duration import Interval
from narwhals._pandas_like.utils import (
    ALIAS_DICT,
    calculate_timestamp_date,
    calculate_timestamp_datetime,
    native_to_narwhals_dtype,
)
from narwhals._utils import Implementation

if TYPE_CHECKING:
    import dask.dataframe.dask_expr as dx

    from narwhals._dask.expr import DaskExpr
    from narwhals.typing import TimeUnit


class DaskExprDateTimeNamespace(
    LazyExprNamespace["DaskExpr"], DateTimeNamespace["DaskExpr"]
):
    def date(self) -> DaskExpr:
        return self.compliant._with_callable(lambda expr: expr.dt.date)

    def year(self) -> DaskExpr:
        return self.compliant._with_callable(lambda expr: expr.dt.year)

    def month(self) -> DaskExpr:
        return self.compliant._with_callable(lambda expr: expr.dt.month)

    def day(self) -> DaskExpr:
        return self.compliant._with_callable(lambda expr: expr.dt.day)

    def hour(self) -> DaskExpr:
        return self.compliant._with_callable(lambda expr: expr.dt.hour)

    def minute(self) -> DaskExpr:
        return self.compliant._with_callable(lambda expr: expr.dt.minute)

    def second(self) -> DaskExpr:
        return self.compliant._with_callable(lambda expr: expr.dt.second)

    def millisecond(self) -> DaskExpr:
        return self.compliant._with_callable(lambda expr: expr.dt.microsecond // 1000)

    def microsecond(self) -> DaskExpr:
        return self.compliant._with_callable(lambda expr: expr.dt.microsecond)

    def nanosecond(self) -> DaskExpr:
        return self.compliant._with_callable(
            lambda expr: expr.dt.microsecond * 1000 + expr.dt.nanosecond
        )

    def ordinal_day(self) -> DaskExpr:
        return self.compliant._with_callable(lambda expr: expr.dt.day_of_year)

    def weekday(self) -> DaskExpr:
        return self.compliant._with_callable(
            lambda expr: expr.dt.day_of_week + 1  # Dask is 0-6
        )

    def to_string(self, format: str) -> DaskExpr:
        return self.compliant._with_callable(
            lambda expr: expr.dt.strftime(format.replace("%.f", ".%f"))
        )

    def replace_time_zone(self, time_zone: str | None) -> DaskExpr:
        return self.compliant._with_callable(
            lambda expr: (
                expr.dt.tz_localize(None).dt.tz_localize(time_zone)
                if time_zone is not None
                else expr.dt.tz_localize(None)
            )
        )

    def convert_time_zone(self, time_zone: str) -> DaskExpr:
        def func(s: dx.Series) -> dx.Series:
            dtype = native_to_narwhals_dtype(
                s.dtype, self.compliant._version, Implementation.DASK
            )
            if dtype.time_zone is None:  # type: ignore[attr-defined]
                return s.dt.tz_localize("UTC").dt.tz_convert(time_zone)  # pyright: ignore[reportAttributeAccessIssue]
            return s.dt.tz_convert(time_zone)  # pyright: ignore[reportAttributeAccessIssue]

        return self.compliant._with_callable(func)

    # ignoring coverage due to https://github.com/narwhals-dev/narwhals/issues/2808.
    def timestamp(self, time_unit: TimeUnit) -> DaskExpr:  # pragma: no cover
        def func(s: dx.Series) -> dx.Series:
            dtype = native_to_narwhals_dtype(
                s.dtype, self.compliant._version, Implementation.DASK
            )
            is_pyarrow_dtype = "pyarrow" in str(dtype)
            mask_na = s.isna()
            dtypes = self.compliant._version.dtypes
            if dtype == dtypes.Date:
                # Date is only supported in pandas dtypes if pyarrow-backed
                s_cast = s.astype("Int32[pyarrow]")
                result = calculate_timestamp_date(s_cast, time_unit)
            elif isinstance(dtype, dtypes.Datetime):
                original_time_unit = dtype.time_unit
                s_cast = (
                    s.astype("Int64[pyarrow]") if is_pyarrow_dtype else s.astype("int64")
                )
                result = calculate_timestamp_datetime(
                    s_cast, original_time_unit, time_unit
                )
            else:
                msg = "Input should be either of Date or Datetime type"
                raise TypeError(msg)
            return result.where(~mask_na)  # pyright: ignore[reportReturnType]

        return self.compliant._with_callable(func)

    def total_minutes(self) -> DaskExpr:
        return self.compliant._with_callable(lambda expr: expr.dt.total_seconds() // 60)

    def total_seconds(self) -> DaskExpr:
        return self.compliant._with_callable(lambda expr: expr.dt.total_seconds() // 1)

    def total_milliseconds(self) -> DaskExpr:
        return self.compliant._with_callable(
            lambda expr: expr.dt.total_seconds() * MS_PER_SECOND // 1
        )

    def total_microseconds(self) -> DaskExpr:
        return self.compliant._with_callable(
            lambda expr: expr.dt.total_seconds() * US_PER_SECOND // 1
        )

    def total_nanoseconds(self) -> DaskExpr:
        return self.compliant._with_callable(
            lambda expr: expr.dt.total_seconds() * NS_PER_SECOND // 1
        )

    def truncate(self, every: str) -> DaskExpr:
        interval = Interval.parse(every)
        unit = interval.unit
        if unit in {"mo", "q", "y"}:
            msg = f"Truncating to {unit} is not yet supported for dask."
            raise NotImplementedError(msg)
        freq = f"{interval.multiple}{ALIAS_DICT.get(unit, unit)}"
        return self.compliant._with_callable(lambda expr: expr.dt.floor(freq))

    def offset_by(self, by: str) -> DaskExpr:
        def func(s: dx.Series) -> dx.Series:
            interval = Interval.parse_no_constraints(by)
            unit = interval.unit
            if unit in {"y", "q", "mo", "d", "ns"}:
                msg = f"Offsetting by {unit} is not yet supported for dask."
                raise NotImplementedError(msg)
            offset = interval.to_timedelta()
            return s.add(offset)

        return self.compliant._with_callable(func)



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING

import narwhals as nw

if TYPE_CHECKING:
    from narwhals.typing import FrameT


def query(customer_ds: FrameT, line_item_ds: FrameT, orders_ds: FrameT) -> FrameT:
    var_1 = var_2 = datetime(1995, 3, 15)
    var_3 = "BUILDING"

    return (
        customer_ds.filter(nw.col("c_mktsegment") == var_3)
        .join(orders_ds, left_on="c_custkey", right_on="o_custkey")
        .join(line_item_ds, left_on="o_orderkey", right_on="l_orderkey")
        .filter(nw.col("o_orderdate") < var_2, nw.col("l_shipdate") > var_1)
        .with_columns(
            (nw.col("l_extendedprice") * (1 - nw.col("l_discount"))).alias("revenue")
        )
        .group_by(["o_orderkey", "o_orderdate", "o_shippriority"])
        .agg([nw.sum("revenue")])
        .select(
            [
                nw.col("o_orderkey").alias("l_orderkey"),
                "revenue",
                "o_orderdate",
                "o_shippriority",
            ]
        )
        .sort(by=["revenue", "o_orderdate"], descending=[True, False])
        .head(10)
    )



from __future__ import annotations

import abc
from functools import lru_cache
from typing import TYPE_CHECKING, Any, ClassVar, Final, Protocol, TypeVar, overload

import polars as pl

from narwhals._duration import Interval
from narwhals._utils import (
    Implementation,
    Version,
    _DeferredIterable,
    _StoresCompliant,
    _StoresNative,
    deep_getattr,
    isinstance_or_issubclass,
)
from narwhals.exceptions import (
    ColumnNotFoundError,
    ComputeError,
    DuplicateError,
    InvalidOperationError,
    NarwhalsError,
    ShapeError,
)

if TYPE_CHECKING:
    from collections.abc import Callable, Iterable, Iterator, Mapping

    from typing_extensions import TypeIs

    from narwhals._compliant.typing import Accessor
    from narwhals._polars.dataframe import Method
    from narwhals._polars.expr import PolarsExpr
    from narwhals._polars.series import PolarsSeries
    from narwhals.dtypes import DType
    from narwhals.typing import IntoDType

    T = TypeVar("T")
    NativeT = TypeVar(
        "NativeT", bound="pl.DataFrame | pl.LazyFrame | pl.Series | pl.Expr"
    )

NativeT_co = TypeVar("NativeT_co", "pl.Series", "pl.Expr", covariant=True)
CompliantT_co = TypeVar("CompliantT_co", "PolarsSeries", "PolarsExpr", covariant=True)
CompliantT = TypeVar("CompliantT", "PolarsSeries", "PolarsExpr")

BACKEND_VERSION = Implementation.POLARS._backend_version()
"""Static backend version for `polars`."""

SERIES_ACCEPTS_PD_INDEX: Final[bool] = BACKEND_VERSION >= (0, 20, 7)
"""`pl.Series(values: pd.Index)` fixed in https://github.com/pola-rs/polars/pull/14087"""

SERIES_RESPECTS_DTYPE: Final[bool] = BACKEND_VERSION >= (0, 20, 26)
"""`pl.Series(dtype=...)` fixed in https://github.com/pola-rs/polars/pull/15962

Includes `SERIES_ACCEPTS_PD_INDEX`.
"""

HAS_INT_128 = BACKEND_VERSION >= (1, 18, 0)
"""https://github.com/pola-rs/polars/pull/20232"""

FROM_DICTS_ACCEPTS_MAPPINGS: Final[bool] = BACKEND_VERSION >= (1, 30, 0)
"""`pl.from_dicts(data: Iterable[Mapping[str, Any]])` since https://github.com/pola-rs/polars/pull/22638"""

HAS_UINT_128 = BACKEND_VERSION >= (1, 34, 0)
"""https://github.com/pola-rs/polars/pull/24346"""


@overload
def extract_native(obj: _StoresNative[NativeT]) -> NativeT: ...
@overload
def extract_native(obj: T) -> T: ...
def extract_native(obj: _StoresNative[NativeT] | T) -> NativeT | T:
    return obj.native if _is_compliant_polars(obj) else obj


def _is_compliant_polars(
    obj: _StoresNative[NativeT] | Any,
) -> TypeIs[_StoresNative[NativeT]]:
    from narwhals._polars.dataframe import PolarsDataFrame, PolarsLazyFrame
    from narwhals._polars.expr import PolarsExpr
    from narwhals._polars.series import PolarsSeries

    return isinstance(obj, (PolarsDataFrame, PolarsLazyFrame, PolarsSeries, PolarsExpr))


def extract_args_kwargs(
    args: Iterable[Any], kwds: Mapping[str, Any], /
) -> tuple[Iterator[Any], dict[str, Any]]:
    it_args = (extract_native(arg) for arg in args)
    return it_args, {k: extract_native(v) for k, v in kwds.items()}


@lru_cache(maxsize=16)
def native_to_narwhals_dtype(  # noqa: C901, PLR0912
    dtype: pl.DataType, version: Version
) -> DType:
    dtypes = version.dtypes
    if dtype == pl.Float64:
        return dtypes.Float64()
    if dtype == pl.Float32:
        return dtypes.Float32()
    if HAS_INT_128 and dtype == pl.Int128:
        return dtypes.Int128()
    if dtype == pl.Int64:
        return dtypes.Int64()
    if dtype == pl.Int32:
        return dtypes.Int32()
    if dtype == pl.Int16:
        return dtypes.Int16()
    if dtype == pl.Int8:
        return dtypes.Int8()
    if HAS_UINT_128 and dtype == pl.UInt128:
        return dtypes.UInt128()
    if dtype == pl.UInt64:
        return dtypes.UInt64()
    if dtype == pl.UInt32:
        return dtypes.UInt32()
    if dtype == pl.UInt16:
        return dtypes.UInt16()
    if dtype == pl.UInt8:
        return dtypes.UInt8()
    if dtype == pl.String:
        return dtypes.String()
    if dtype == pl.Boolean:
        return dtypes.Boolean()
    if dtype == pl.Object:
        return dtypes.Object()
    if dtype == pl.Categorical:
        return dtypes.Categorical()
    if isinstance_or_issubclass(dtype, pl.Enum):
        if version is Version.V1:
            return dtypes.Enum()  # type: ignore[call-arg]
        categories = _DeferredIterable(dtype.categories.to_list)
        return dtypes.Enum(categories)
    if dtype == pl.Date:
        return dtypes.Date()
    if isinstance_or_issubclass(dtype, pl.Datetime):
        return (
            dtypes.Datetime()
            if dtype is pl.Datetime
            else dtypes.Datetime(dtype.time_unit, dtype.time_zone)
        )
    if isinstance_or_issubclass(dtype, pl.Duration):
        return (
            dtypes.Duration()
            if dtype is pl.Duration
            else dtypes.Duration(dtype.time_unit)
        )
    if isinstance_or_issubclass(dtype, pl.Struct):
        fields = [
            dtypes.Field(name, native_to_narwhals_dtype(tp, version))
            for name, tp in dtype
        ]
        return dtypes.Struct(fields)
    if isinstance_or_issubclass(dtype, pl.List):
        return dtypes.List(native_to_narwhals_dtype(dtype.inner, version))
    if isinstance_or_issubclass(dtype, pl.Array):
        outer_shape = dtype.width if BACKEND_VERSION < (0, 20, 30) else dtype.size
        return dtypes.Array(native_to_narwhals_dtype(dtype.inner, version), outer_shape)
    if isinstance_or_issubclass(dtype, pl.Decimal):
        return dtypes.Decimal(dtype.precision, dtype.scale)
    if dtype == pl.Time:
        return dtypes.Time()
    if dtype == pl.Binary:
        return dtypes.Binary()
    return dtypes.Unknown()


dtypes = Version.MAIN.dtypes


def _version_dependent_dtypes() -> dict[type[DType], pl.DataType]:
    if not HAS_INT_128:  # pragma: no cover
        return {}
    nw_to_pl: dict[type[DType], pl.DataType] = {dtypes.Int128: pl.Int128()}
    return nw_to_pl | {dtypes.UInt128: pl.UInt128()} if HAS_UINT_128 else nw_to_pl


NW_TO_PL_DTYPES: Mapping[type[DType], pl.DataType] = {
    dtypes.Float64: pl.Float64(),
    dtypes.Float32: pl.Float32(),
    dtypes.Binary: pl.Binary(),
    dtypes.String: pl.String(),
    dtypes.Boolean: pl.Boolean(),
    dtypes.Categorical: pl.Categorical(),
    dtypes.Date: pl.Date(),
    dtypes.Time: pl.Time(),
    dtypes.Int8: pl.Int8(),
    dtypes.Int16: pl.Int16(),
    dtypes.Int32: pl.Int32(),
    dtypes.Int64: pl.Int64(),
    dtypes.UInt8: pl.UInt8(),
    dtypes.UInt16: pl.UInt16(),
    dtypes.UInt32: pl.UInt32(),
    dtypes.UInt64: pl.UInt64(),
    dtypes.Object: pl.Object(),
    dtypes.Unknown: pl.Unknown(),
    **_version_dependent_dtypes(),
}


def narwhals_to_native_dtype(  # noqa: C901
    dtype: IntoDType, version: Version
) -> pl.DataType:
    dtypes = version.dtypes
    base_type = dtype.base_type()
    if pl_type := NW_TO_PL_DTYPES.get(base_type):
        return pl_type
    if isinstance_or_issubclass(dtype, dtypes.Enum):
        if version is Version.V1:
            msg = "Converting to Enum is not supported in narwhals.stable.v1"
            raise NotImplementedError(msg)
        if isinstance(dtype, dtypes.Enum):
            return pl.Enum(dtype.categories)
        msg = "Can not cast / initialize Enum without categories present"
        raise ValueError(msg)
    if isinstance_or_issubclass(dtype, dtypes.Datetime):
        return pl.Datetime(dtype.time_unit, dtype.time_zone)  # type: ignore[arg-type]
    if isinstance_or_issubclass(dtype, dtypes.Duration):
        return pl.Duration(dtype.time_unit)  # type: ignore[arg-type]
    if isinstance_or_issubclass(dtype, dtypes.List):
        return pl.List(narwhals_to_native_dtype(dtype.inner, version))
    if isinstance_or_issubclass(dtype, dtypes.Struct):
        fields = [
            pl.Field(field.name, narwhals_to_native_dtype(field.dtype, version))
            for field in dtype.fields
        ]
        return pl.Struct(fields)
    if isinstance_or_issubclass(dtype, dtypes.Array):  # pragma: no cover
        size = dtype.size
        kwargs = {"width": size} if BACKEND_VERSION < (0, 20, 30) else {"shape": size}
        return pl.Array(narwhals_to_native_dtype(dtype.inner, version), **kwargs)
    if isinstance_or_issubclass(dtype, dtypes.Decimal):
        return pl.Decimal(dtype.precision, dtype.scale)
    return pl.Unknown()  # pragma: no cover


def _is_polars_exception(exception: Exception) -> bool:
    if BACKEND_VERSION >= (1,):
        # Old versions of Polars didn't have PolarsError.
        return isinstance(exception, pl.exceptions.PolarsError)
    # Last attempt, for old Polars versions.
    return "polars.exceptions" in str(type(exception))  # pragma: no cover


def _is_cudf_exception(exception: Exception) -> bool:
    # These exceptions are raised when running polars on GPUs via cuDF
    return str(exception).startswith("CUDF failure")


def catch_polars_exception(exception: Exception) -> NarwhalsError | Exception:
    if isinstance(exception, pl.exceptions.ColumnNotFoundError):
        return ColumnNotFoundError(str(exception))
    if isinstance(exception, pl.exceptions.ShapeError):
        return ShapeError(str(exception))
    if isinstance(exception, pl.exceptions.InvalidOperationError):
        return InvalidOperationError(str(exception))
    if isinstance(exception, pl.exceptions.DuplicateError):
        return DuplicateError(str(exception))
    if isinstance(exception, pl.exceptions.ComputeError):
        return ComputeError(str(exception))
    if _is_polars_exception(exception) or _is_cudf_exception(exception):
        return NarwhalsError(str(exception))  # pragma: no cover
    # Just return exception as-is.
    return exception


class PolarsAnyNamespace(
    _StoresCompliant[CompliantT_co],
    _StoresNative[NativeT_co],
    Protocol[CompliantT_co, NativeT_co],
):
    _accessor: ClassVar[Accessor]

    def __getattr__(self, attr: str) -> Callable[..., CompliantT_co]:
        def func(*args: Any, **kwargs: Any) -> CompliantT_co:
            pos, kwds = extract_args_kwargs(args, kwargs)
            method = deep_getattr(self.native, self._accessor, attr)
            return self.compliant._with_native(method(*pos, **kwds))

        return func


class PolarsDateTimeNamespace(PolarsAnyNamespace[CompliantT, NativeT_co]):
    _accessor: ClassVar[Accessor] = "dt"

    def truncate(self, every: str) -> CompliantT:
        # Ensure consistent error message is raised.
        Interval.parse(every)
        return self.__getattr__("truncate")(every)

    def offset_by(self, by: str) -> CompliantT:
        # Ensure consistent error message is raised.
        Interval.parse_no_constraints(by)
        return self.__getattr__("offset_by")(by)

    to_string: Method[CompliantT]
    replace_time_zone: Method[CompliantT]
    convert_time_zone: Method[CompliantT]
    timestamp: Method[CompliantT]
    date: Method[CompliantT]
    year: Method[CompliantT]
    month: Method[CompliantT]
    day: Method[CompliantT]
    hour: Method[CompliantT]
    minute: Method[CompliantT]
    second: Method[CompliantT]
    millisecond: Method[CompliantT]
    microsecond: Method[CompliantT]
    nanosecond: Method[CompliantT]
    ordinal_day: Method[CompliantT]
    weekday: Method[CompliantT]
    total_minutes: Method[CompliantT]
    total_seconds: Method[CompliantT]
    total_milliseconds: Method[CompliantT]
    total_microseconds: Method[CompliantT]
    total_nanoseconds: Method[CompliantT]


class PolarsStringNamespace(PolarsAnyNamespace[CompliantT, NativeT_co]):
    _accessor: ClassVar[Accessor] = "str"

    # NOTE: Use `abstractmethod` if we have defs to implement, but also `Method` usage
    @abc.abstractmethod
    def to_titlecase(self) -> CompliantT: ...

    @abc.abstractmethod
    def zfill(self, width: int) -> CompliantT: ...

    len_chars: Method[CompliantT]
    replace: Method[CompliantT]
    replace_all: Method[CompliantT]
    strip_chars: Method[CompliantT]
    starts_with: Method[CompliantT]
    ends_with: Method[CompliantT]
    contains: Method[CompliantT]
    slice: Method[CompliantT]
    split: Method[CompliantT]
    to_date: Method[CompliantT]
    to_datetime: Method[CompliantT]
    to_lowercase: Method[CompliantT]
    to_uppercase: Method[CompliantT]
    pad_start: Method[CompliantT]
    pad_end: Method[CompliantT]


class PolarsCatNamespace(PolarsAnyNamespace[CompliantT, NativeT_co]):
    _accessor: ClassVar[Accessor] = "cat"
    get_categories: Method[CompliantT]


class PolarsListNamespace(PolarsAnyNamespace[CompliantT, NativeT_co]):
    _accessor: ClassVar[Accessor] = "list"

    @abc.abstractmethod
    def len(self) -> CompliantT: ...

    get: Method[CompliantT]

    unique: Method[CompliantT]

    max: Method[CompliantT]

    mean: Method[CompliantT]

    median: Method[CompliantT]

    min: Method[CompliantT]

    sort: Method[CompliantT]

    sum: Method[CompliantT]


class PolarsStructNamespace(PolarsAnyNamespace[CompliantT, NativeT_co]):
    _accessor: ClassVar[Accessor] = "struct"
    field: Method[CompliantT]



"""Script to automatically sort members lists in API reference markdown files."""

from __future__ import annotations

import re
import sys
from pathlib import Path


def sort_members_in_markdown(file_path: Path) -> int:
    """Sort members lists in a markdown file alphabetically.

    Returns:
        1 if the file was modified, 0 if no changes were needed.
    """
    content = file_path.read_text(encoding="utf-8")

    # Pattern matches "members:" followed by list items (lines starting with "- ")
    pattern = r"(members:)((?:\n\s+-[^\n]+)+)"

    def sort_list(match: re.Match[str]) -> str:
        """Sort a matched members list section."""
        prefix = match.group(1)  # "members:"
        items = match.group(2)  # All list items with newlines and indentation

        item_pattern = r"(\n\s+-)([^\n]+)"  # Extract individual items: (\n + spaces + -) and (content)
        matches = re.findall(item_pattern, items)

        # Sort alphabetically by the item content (excluding indentation and dash)
        sorted_items = sorted(matches, key=lambda x: x[1].strip())

        # Reconstruct: "members:" + sorted items with their indentation preserved
        return prefix + "".join(f"{indent}{content}" for indent, content in sorted_items)

    sorted_content = re.sub(pattern, sort_list, content)
    if sorted_content == content:
        return 0

    file_path.write_text(sorted_content, encoding="utf-8")
    print(f"Sorting members in {file_path}")
    return 1


PATH = Path("docs") / "api-reference"
FILES_TO_SKIP = {"dtypes", "typing"}

ret = max(
    sort_members_in_markdown(file_path=file_path)
    for file_path in PATH.glob("*.md")
    if file_path.stem not in FILES_TO_SKIP
)

sys.exit(ret)



from __future__ import annotations

from datetime import datetime

import narwhals as nw
from tests.utils import Constructor, assert_equal_data


def test_unpivot(constructor: Constructor) -> None:
    data = {
        "date": [datetime(2020, 1, 1), datetime(2020, 1, 2), datetime(2020, 1, 3)],
        "aapl": [110, 100, 105],
        "tsla": [220, 200, 210],
        "msft": [330, 300, 315],
        "nflx": [420, 400, 440],
    }
    df = nw.from_native(constructor(data))
    result = df.unpivot(index="date", value_name="price").sort(by=["date", "variable"])
    expected = {
        "date": [
            *[datetime(2020, 1, 1)] * 4,
            *[datetime(2020, 1, 2)] * 4,
            *[datetime(2020, 1, 3)] * 4,
        ],
        "variable": [*["aapl", "msft", "nflx", "tsla"] * 3],
        "price": [110, 330, 420, 220, 100, 300, 400, 200, 105, 315, 440, 210],
    }
    assert_equal_data(result, expected)



from __future__ import annotations

from typing import TYPE_CHECKING

import narwhals as nw

if TYPE_CHECKING:
    from narwhals.typing import FrameT


def query(customer_ds: FrameT, orders_ds: FrameT) -> FrameT:
    q1 = (
        customer_ds.with_columns(nw.col("c_phone").str.slice(0, 2).alias("cntrycode"))
        .filter(nw.col("cntrycode").str.contains("13|31|23|29|30|18|17"))
        .select("c_acctbal", "c_custkey", nw.col("cntrycode").cast(nw.Int64()))
    )

    q2 = q1.filter(nw.col("c_acctbal") > 0.0).select(
        nw.col("c_acctbal").mean().alias("avg_acctbal")
    )

    q3 = (
        orders_ds.select("o_custkey")
        .unique("o_custkey")
        .with_columns(nw.col("o_custkey").alias("c_custkey"))
    )

    return (
        q1.join(q3, left_on="c_custkey", right_on="c_custkey", how="left")  # pyright: ignore[reportArgumentType]
        .filter(nw.col("o_custkey").is_null())
        .join(q2, how="cross")  # pyright: ignore[reportArgumentType]
        .filter(nw.col("c_acctbal") > nw.col("avg_acctbal"))
        .group_by("cntrycode")
        .agg(
            nw.col("c_acctbal").count().alias("numcust"),
            nw.col("c_acctbal").sum().alias("totacctbal"),
        )
        .sort("cntrycode")
    )



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

import narwhals as nw
from tests.utils import PANDAS_VERSION, assert_equal_data

if TYPE_CHECKING:
    from tests.utils import Constructor, ConstructorEager

data = {"a": [[3, None, 2, 2, 4, None], [-1], None, [None, None, None], []]}
expected = [2, -1, None, None, None]


def test_min_expr(request: pytest.FixtureRequest, constructor: Constructor) -> None:
    if any(backend in str(constructor) for backend in ("dask", "cudf")):
        request.applymarker(pytest.mark.xfail)
    if "pandas" in str(constructor):
        if PANDAS_VERSION < (2, 2):
            pytest.skip()
        pytest.importorskip("pyarrow")
    result = nw.from_native(constructor(data)).select(
        nw.col("a").cast(nw.List(nw.Int32())).list.min()
    )
    assert_equal_data(result, {"a": expected})


def test_min_series(
    request: pytest.FixtureRequest, constructor_eager: ConstructorEager
) -> None:
    if any(backend in str(constructor_eager) for backend in ("cudf",)):
        request.applymarker(pytest.mark.xfail)
    if "pandas" in str(constructor_eager):
        if PANDAS_VERSION < (2, 2):
            pytest.skip()
        pytest.importorskip("pyarrow")
    df = nw.from_native(constructor_eager(data), eager_only=True)
    result = df["a"].cast(nw.List(nw.Int32())).list.min().to_list()
    assert_equal_data({"a": result}, {"a": expected})



from __future__ import annotations

from contextlib import suppress
from importlib.util import find_spec
from typing import TYPE_CHECKING

import pytest

from tpch import constants
from tpch.classes import Backend, Query, Query as q  # noqa: N813

if TYPE_CHECKING:
    from collections.abc import Iterator

    from tpch.typing_ import ScaleFactor


def is_xdist_worker(obj: pytest.FixtureRequest | pytest.Config, /) -> bool:
    # Adapted from https://github.com/pytest-dev/pytest-xdist/blob/8b60b1ef5d48974a1cb69bc1a9843564bdc06498/src/xdist/plugin.py#L337-L349
    return hasattr(obj if isinstance(obj, pytest.Config) else obj.config, "workerinput")


def pytest_configure(config: pytest.Config) -> None:
    """Generate TPC-H data if it doesn't exist for the requested scale factor.

    [`configure`] runs after `addoption`, ensuring data is available before test collection.

    [`configure`]: https://docs.pytest.org/en/stable/reference/reference.html#pytest.hookspec.pytest_configure
    """
    # Only run before the session starts, instead of 1 + (`--numprocesses`)
    if is_xdist_worker(config):
        return
    from tpch.generate_data import TPCHGen

    TPCHGen.from_pytest(config).run()


def pytest_addoption(parser: pytest.Parser) -> None:
    from tests.conftest import DEFAULT_CONSTRUCTORS

    parser.addoption(
        "--constructors",
        default=DEFAULT_CONSTRUCTORS,
        help="<sink for defaults in VSC getting injected>",
    )
    parser.addoption(
        "--scale-factor",
        default=constants.SCALE_FACTOR_DEFAULT,
        choices=constants.SCALE_FACTORS,
        help="TPC-H scale factor to use for tests (default: %(default)s)",
    )


def iter_backends() -> Iterator[Backend]:
    yield Backend("polars[lazy]")
    if find_spec("pyarrow"):
        yield Backend("pyarrow")
        if find_spec("pandas"):
            import pandas as pd

            # These options are deprecated in pandas >= 3.0 but needed for older versions
            with suppress(Exception):
                pd.options.mode.copy_on_write = True
            with suppress(Exception):
                pd.options.future.infer_string = True  # pyright: ignore[reportAttributeAccessIssue, reportOptionalMemberAccess]
            yield Backend("pandas[pyarrow]", engine="pyarrow", dtype_backend="pyarrow")
        if find_spec("dask") and find_spec("dask.dataframe"):
            yield Backend("dask", engine="pyarrow", dtype_backend="pyarrow")
    if find_spec("duckdb"):
        yield Backend("duckdb")
        if find_spec("sqlframe"):
            from sqlframe.duckdb import DuckDBSession

            yield Backend("sqlframe", session=DuckDBSession())


@pytest.fixture(params=iter_backends(), ids=repr)
def backend(request: pytest.FixtureRequest) -> Backend:
    return request.param


def iter_queries() -> Iterator[Query]:
    yield from (
        q("q1", "lineitem"),
        q("q2", "region", "nation", "supplier", "part", "partsupp"),
        q("q3", "customer", "lineitem", "orders"),
        q("q4", "lineitem", "orders"),
        q("q5", "region", "nation", "customer", "lineitem", "orders", "supplier"),
        q("q6", "lineitem"),
        q("q7", "nation", "customer", "lineitem", "orders", "supplier"),
        q("q8", "part", "supplier", "lineitem", "orders", "customer", "nation", "region"),
        q("q9", "part", "partsupp", "nation", "lineitem", "orders", "supplier"),
        q("q10", "customer", "nation", "lineitem", "orders"),
        q("q11", "nation", "partsupp", "supplier"),
        q("q12", "lineitem", "orders"),
        q("q13", "customer", "orders"),
        q("q14", "lineitem", "part"),
        q("q15", "lineitem", "supplier"),
        q("q16", "part", "partsupp", "supplier"),
        q("q17", "lineitem", "part"),
        q("q18", "customer", "lineitem", "orders"),
        q("q19", "lineitem", "part"),
        q("q20", "part", "partsupp", "nation", "lineitem", "supplier"),
        q("q21", "lineitem", "nation", "orders", "supplier"),
        q("q22", "customer", "orders"),
    )


@pytest.fixture(scope="session")
def scale_factor(request: pytest.FixtureRequest) -> ScaleFactor:
    return request.config.getoption("--scale-factor")


@pytest.fixture(params=iter_queries(), ids=repr)
def query(request: pytest.FixtureRequest, scale_factor: ScaleFactor) -> Query:
    result: Query = request.param
    return result.with_scale_factor(scale_factor)
